#!/usr/bin/env python
##!/bin/env python
##!/usr/bin/python

import pysam
import sys
import os.path
import time
import resource

###########################################################################################################
#
#  pibase - GET pOSITION iNFORMATION AT base POSITION OF INTEREST
#
###########################################################################################################

def setlimits():
	#############################################################################
	# sets resource limits (mandatory for job interoperability on cluster)
	# to 10MB Stack + 1GB data + 1GB RLIMIT_RSS
	# first steps taken from
	# http://docs.python.org/library/resource.html and
	# http://www.doughellmann.com/PyMOTW/resource/
	# Author: MICHAEL FORSTER m.forster@ikmb.uni-kiel.de
	# Version: 13 September 2010
	#############################################################################

	# usage = resource.getrusage(resource.RUSAGE_SELF)

	#print ' usage ='
	#print usage
	
	#for name, desc in [
	#('RLIMIT_CORE', 'core file size'),
	#('RLIMIT_CPU',  'CPU time'),
	#('RLIMIT_FSIZE', 'file size'),
	#('RLIMIT_DATA', 'heap size'),
	#('RLIMIT_STACK', 'stack size'),
	#('RLIMIT_RSS', 'resident set size'),
	#('RLIMIT_NPROC', 'number of processes'),
	#('RLIMIT_NOFILE', 'number of open files'),
	#('RLIMIT_MEMLOCK', 'lockable memory address'),
	#]:
	#	limit_num = getattr(resource, name)
	#	soft, hard = resource.getrlimit(limit_num)
	#	print 'Maximum %-25s (%-15s) : %20s %20s' % (desc, name, soft, hard)

	#############################################################################
	# Adjust the maximum size (in bytes) of the process's heap
	#############################################################################

	soft, hard = resource.getrlimit(resource.RLIMIT_DATA)
	#print 'Data/heap limit starts as  :', soft

	# 1GB heap:
	newheap=1024*1024*1024

	resource.setrlimit(resource.RLIMIT_DATA, (newheap, hard))

	soft, hard = resource.getrlimit(resource.RLIMIT_DATA)
	#print 'Data/heap limit changed to :', soft

	#############################################################################
	# ca 10MB original stack changed to exactly 10MB stack:
	#############################################################################

	#newstack=10*1024*1024+1
	newstack=10*1024*1024
	soft, hard = resource.getrlimit(resource.RLIMIT_STACK)
	#print 'Stack limit starts as  :', soft
	resource.setrlimit(resource.RLIMIT_STACK, (newstack, hard))
	soft, hard = resource.getrlimit(resource.RLIMIT_STACK)
	#print 'Stack limit changed to :', soft

	#############################################################################
	# 1GB RLIMIT_RSS
	# "Maximale resident set size in Byte. Bei Speicherengpaessen entzieht der Kern
	# den Prozessen, die den resident set size ueberschreiten, den zuviel
	# angeforderten Speicher."
	# http://pronix.linuxdelta.de/C/Linuxprogrammierung/Linuxsystemprogrammieren_C_Kurs_Kapitel3a.shtml
	#############################################################################

	newrss=1024*1024*1024
	soft, hard = resource.getrlimit(resource.RLIMIT_RSS)
	#print 'RLIMIT_RSS starts as  :', soft
	resource.setrlimit(resource.RLIMIT_RSS, (newrss, hard))
	soft, hard = resource.getrlimit(resource.RLIMIT_RSS)
	#print 'RLIMIT_RSS changed to :', soft

	#sys.exit()


def pib (version, details, verbose, onechr, onepoi, refseqfile, tablefile, bamfile, outfile, LFR, \
         qvmin, MLmin, MMmax, RQVmin, rmax):
	'''        #######################################################################################
	#  pibase_bamref v1.4.7 - GET pOSITION iNFORMATION AT base POSITION OF INTEREST
	#######################################################################################
	#  pibase    [-[v][d][p chr]] <poi or table> ref bamfile outfile LR
	#            [[QVmin [[MLmin] [[MMmax] [RQVmin] [maxr]]]]]
	#
	#            []=optional
	#
	#   -vdp     [v]erbose output into terminal,
	#            [d]etailed list of reads,
	#            at single [p]osition: <chr> <poi> 
	#
	#   chr      chromosome name (or contig name)
	#   poi      position of interest (chromosomal coordinate, 1-based)
	#
	#   table    file of poi's: vcf, gff3, pileup, 
	#            or plain tab-file: <chr> TAB <poi> LINEBREAK
	#            The format is detected by pibase_bamref. If the plain tab-file is used,
	#            the lines must be sorted by chr-names, and the chr-names must be ordered
	#            in the sequence of the reference sequence file or BAM-file-header.
	#
	#   ref      reference sequence file
	#   bamfile  sorted bam file, bam index file must also exist
	#   outfile  name for (tab-separated) output file
	#   LR       max length of reads (e.g. 50)
	#
	#   QVmin    optional min base quality @Filter1+ (default: 20)
	#   MLmin    optional min mapped read length @Filter2+ (default: 49)
	#   MMmax    optional max number of base mismatches in read @Filter3+ (default: 1)
	#   RQVmin   optional min read mapping quality @Filter4 (default: 20)
	#   maxr     optional max number of reads considered at a poi (default: -1 unlimited)
	#######################################################################################
	# Notes:
	# 1 pibase cannot detect whether the BAM file is the original complete file, or just 
	#   a part of the original file.
	# 2 For the most complete information: give pibase the original file (which should 
	#   include the ambiguously mapped reads, and which can also include "duplicate" reads)
	# 3 pibase needs sorted indexed BAM files with MD-tags, e.g. created using samtools:
	#     samtools sort
	#     samtools calmd
	#     samtools index
	#######################################################################################
	#  Michael Forster, m.forster@ikmb.uni-kiel.de                            10 Jan 2013
	#  Acknowledgement: Ingo Thomsen (pysam intro, example code, support and discussions)
	#######################################################################################
	'''

	#######################################################################################
	# open bamfile and output file
	#######################################################################################

	#print onepoi
	
	samfile = pysam.Samfile( bamfile , "rb")
	out = open(outfile,"w")

	#######################################################################################
	# get reference sequence chromosomes (or contigs) and their lengths from sam file header
	#######################################################################################

	refs = samfile.references
	nrefs=len(refs)
	lengths = samfile.lengths
	# print refs
	# print lengths

	#######################################################################################
	# Get chromosome (or contig) name, and 0-based position of interest (poi):
	#######################################################################################

	if onepoi > 0:
		
		###################################################################################
		# get single position of interest from command line argument
		###################################################################################

		# For start: specify the position of interest (poi), 0-based (because user entered poi 1-based).
		# For end: specify the poi + 1
		# (required because python for-loop does not execute the j=end'th cycle
		# and will not do anything if end=start)
		poi=onepoi-1
		chrom=onechr
		if onechr in refs:
			ichrom = refs.index(onechr)
		else:
			print '*** Error: chromosome/contig %s is unknown in BAM file.\n' % (onechr)
			sys.exit()
		
	else:

		###################################################################################
		# read multiple positions of interest from input file
		###################################################################################

		# identify table file type:
		tabfiletype=0
		bioscope_gff3=False
		bioscope_diBayes=False
		sampileup=False
		vcfformat=False
		k=0
		tabf=open (tablefile, 'r')
		while True:
			line=tabf.readline()
			line=line.strip()
			if line.startswith('#'):
				if k==0:
					if line.find('gff-version 3')>-1:
						bioscope_gff3=True
					elif line.find('##fileformat=VCF')>-1:
						vcfformat=True
						tabfiletype=3
					k+=1
			elif len(line)==0:
				sys.exit('*** Error: No useable lines in file:  %s' % (tablefile))
			else:
				if tabfiletype!=3:			
					if line.find('SOLiD_diBayes')>-1:
						bioscope_diBayes=True
						if bioscope_gff3:
							tabfiletype=1
					else:
						cols=line.split()
						if len(cols) > 2 or line.find('chr')==0:
							# assume pileup format or first two columns of pileup format:
							sampileup=True
							tabfiletype=3
						else:
							# assume tab-separated chromosome-name / coordinate:
							tabfiletype=2
				break
		# Get chromosome number and position
		
		if tabfiletype==0:
			sys.exit('*** Error: Bad gff3 format in file:  %s' % (tablefile))
		elif tabfiletype==1:
			# Get ichrom (0-based) and poi (0-based) from Bioscope diBayes gff3 file:
			try:
				cols = line.split('\t')
				chromn = cols[0][3:]
				ichrom = int(chromn.strip()) - 1
				poi = int(cols[3].strip()) - 1
			except:
				sys.exit('*** Error: Bad gff3 format in first SNP-line of file:  %s\n***        Line is:\n%s\n' % (tablefile,line))
		elif tabfiletype==3:
			# Get ichrom (0-based) and poi (0-based) from vcf file or from samtools pileup file:
			try:
				cols = line.split('\t')
				# Pre-v1.4.2:
				# chromn = cols[0][3:]
				# ichrom = int(chromn.strip()) - 1
				# v1.4.2: Find chromosome name in bam-file header:
				chromn = cols[0].strip()
				if chromn in refs:
					ichrom = refs.index(chromn)
				else:
					print '*** Error: chromosome/contig %s is unknown in BAM file.\n***        Line is:\n%s\n' % (chromn,line)
					sys.exit()
				poi = int(cols[1].strip()) - 1
			except:
				if vcfformat:
					sys.exit('*** Error: Bad vcf format in first SNP-line of file:  %s\n***        Line is:\n%s\n' % (tablefile,line))
				else:
					sys.exit('*** Error: Bad samtools pileup format in first SNP-line of file:  %s\n***        Line is:\n%s\n' % (tablefile,line))
		else:
			# Get ichrom (0-based) and poi (0-based) from tab-separated file:
			try:
				cols = line.split('\t')
				chromn = cols[0].strip()
				if chromn in refs:
					ichrom = refs.index(chromn)
				else:
					print '*** Error: chromosome/contig %s is unknown in BAM file.\n***        Line is:\n%s\n' % (chromn,line)
					sys.exit()
				poi = int(cols[1].strip()) - 1
			except:
				print '*** Error: Bad format in first SNP-line of file:  %s\n***        Line is:\n%s\n' % (tablefile,line)
				sys.exit('***        Line should be: <chromosome_name> TAB <1_based_position_in_chrom>')

		# Translate chromosome number into chromosome name:
		if ichrom+1 > nrefs:
			if nrefs == 1:
				sys.exit('*** Error: Bad <chr#> = %s in file, because only 1 chrom/contig is defined in bamfile header.\nLine:\n%s\n' % (ichrom+1, line))
			else:
				sys.exit('*** Error: Bad <chr#> = %s in file, because only %s chroms/contigs are defined in bamfile header.\nLine:\n%s\n' % (ichrom+1, nrefs, line))
		chrom = refs[ichrom]

	ichrommax = ichrom

	# print '###### refs ##########'
	# print refs[0]
	# print refs
	# print '###### lengths ##########'
	# print lengths
	# sys.exit()

	#######################################################################################
	# open refseq file - pysam does not work as tried (empty) - so doityourself in Python
	#######################################################################################

	# refseq = pysam.Fastafile (refseqfile)
	# fastaseq = refseq.fetch (chrom, 0, 5)
	# fastaseq = refseq.fetch (chrom)
	# print fastaseq[10]
	# sys.exit()
	
	# read chromnames, lengths and binary starts (0-based) from fasta-index-file:
	faifile=refseqfile+'.fai'
	faifh=open(faifile)
	fai={}
	for line in faifh:
		cols=line.split('\t')
		fai[cols[0]]=(int(cols[1]),int(cols[2]))
		# fai.append((cols[0],cols[1],cols[2]))
	# print fai
	# l,j=fai['CHR1']
	# print l,j
	# sys.exit()

	#######################################################################################################
	# create and initialise lists used for unique start point statistics
	#######################################################################################################

	# This following code block is a bit quick and dirty, but:
	#   + much easier to understand,
	#   + much easier to write,
	#   + much easier to test and debug,
	#   + faster to run
	#   than a 3-dimensional array with nested loops

	# count unique start points - initialise lists
	# for ACGTN, for strands 1 and 2, and for
	#    - all bases and
	#    - QV-limited bases and
	#    - QV-limited bases and mapped-length-of-read threshold and
	#    - QV-limited bases, mapped-length-of-read, and mismatch-in-read threshold
	#    - QV-limited bases, mapped-length-of-read, mismatch-in-read, and read-mapping QV threshold
		#
	isa1=[]
	isa2=[]
	isqa1=[]
	isqa2=[]
	isma1=[]
	isma2=[]
	ismma1=[]
	ismma2=[]
	ismqa1=[]
	ismqa2=[]

	isc1=[]
	isc2=[]
	isqc1=[]
	isqc2=[]
	ismc1=[]
	ismc2=[]
	ismmc1=[]
	ismmc2=[]
	ismqc1=[]
	ismqc2=[]

	isg1=[]
	isg2=[]
	isqg1=[]
	isqg2=[]
	ismg1=[]
	ismg2=[]
	ismmg1=[]
	ismmg2=[]
	ismqg1=[]
	ismqg2=[]

	ist1=[]
	ist2=[]
	isqt1=[]
	isqt2=[]
	ismt1=[]
	ismt2=[]
	ismmt1=[]
	ismmt2=[]
	ismqt1=[]
	ismqt2=[]

	isn1=[]
	isn2=[]
	isqn1=[]
	isqn2=[]
	ismn1=[]
	ismn2=[]
	ismmn1=[]
	ismmn2=[]
	ismqn1=[]
	ismqn2=[]
	for j in range(0,LFR):
		isa1.append(int(0))
		isa2.append(int(0))
		isqa1.append(int(0))
		isqa2.append(int(0))
		isma1.append(int(0))
		isma2.append(int(0))
		ismma1.append(int(0))
		ismma2.append(int(0))
		ismqa1.append(int(0))
		ismqa2.append(int(0))

		isc1.append(int(0))
		isc2.append(int(0))
		isqc1.append(int(0))
		isqc2.append(int(0))
		ismc1.append(int(0))
		ismc2.append(int(0))
		ismmc1.append(int(0))
		ismmc2.append(int(0))
		ismqc1.append(int(0))
		ismqc2.append(int(0))

		isg1.append(int(0))
		isg2.append(int(0))
		isqg1.append(int(0))
		isqg2.append(int(0))
		ismg1.append(int(0))
		ismg2.append(int(0))
		ismmg1.append(int(0))
		ismmg2.append(int(0))
		ismqg1.append(int(0))
		ismqg2.append(int(0))

		ist1.append(int(0))
		ist2.append(int(0))
		isqt1.append(int(0))
		isqt2.append(int(0))
		ismt1.append(int(0))
		ismt2.append(int(0))
		ismmt1.append(int(0))
		ismmt2.append(int(0))
		ismqt1.append(int(0))
		ismqt2.append(int(0))

		isn1.append(int(0))
		isn2.append(int(0))
		isqn1.append(int(0))
		isqn2.append(int(0))
		ismn1.append(int(0))
		ismn2.append(int(0))
		ismmn1.append(int(0))
		ismmn2.append(int(0))
		ismqn1.append(int(0))
		ismqn2.append(int(0))

	#######################################################################################################
	# Create and initialise lists for classifying position-of-interest context
	#######################################################################################################

	poicls = []
	poiclt = []
	# dummy records, so that #0, #1 do not happen (#1 is the "normal" classification, #2+ the problems)
	poicls.append('1')
	poicls.append('2')
	poiclt.append('1')
	poiclt.append('2')
	# 4 repeats of single base
	poicls.append('AAAA')
	poicls.append('CCCC')
	poicls.append('GGGG')
	poicls.append('TTTT')
	poicls.append('NNNN')
	# 5 repeats of single base
	poicls.append('AAAAA')
	poicls.append('CCCCC')
	poicls.append('GGGGG')
	poicls.append('TTTTT')
	poicls.append('NNNNN')
	# 3 short tandem repeats
	poiclt.append('ACACAC')
	poiclt.append('AGAGAG')
	poiclt.append('ATATAT')
	poiclt.append('CACACA')
	poiclt.append('CGCGCG')
	poiclt.append('CTCTCT')
	poiclt.append('TATATA')
	poiclt.append('TCTCTC')
	poiclt.append('TGTGTG')
	# 4 short tandem repeats
	poiclt.append('ACACACCA')
	poiclt.append('AGAGAGAG')
	poiclt.append('ATATATAT')
	poiclt.append('CACACACA')
	poiclt.append('CGCGCGCG')
	poiclt.append('CTCTCTCT')
	poiclt.append('TATATATA')
	poiclt.append('TCTCTCTC')
	poiclt.append('TGTGTGTG')

	#######################################################################################################
	#  Write HEADER of OUT FILE
	#######################################################################################################

	#out.write ('# Position Info (1-based coordinates) generated by pibase_bamref v1.4.3 from file '+ bamfile+'\n')
	out.write ('# Position Info (1-based coordinates) generated by pibase_bamref v%s from file %s\n' % (version, bamfile) )
	
	out.write ('# Reference sequence file: '+ refseqfile+'\n')
	out.write ('# Mapping of pibase-chromosome-number to chromosome/contig name:\n')
	out.write ('# \tC\tChr\n')
	icc=0
	for cc in refs:
		icc += 1
		out.write ('# \t%s\t%s\n' % (icc, cc))
	
	out.write ('# Positions of interest taken from file: '+ tablefile+'\n')
	
	outtime = time.strftime("%Y_%m_%d__%H:%M:%S") + "\n"
	
	out.write ('# Max fragment length considered: %s bases\n' % (LFR))
	out.write ('# Base quality value threshold QVmin= '+ str(qvmin)+'\n')
	out.write ('# Mapped-length-of-read threshold MLmin= '+ str(MLmin)+' bases\n')
	out.write ('# Mismatch-in-read threshold MMmax= '+ str(MMmax)+'\n')
	out.write ('# Read-mapping quality value threshold RQVmin= '+ str(RQVmin)+'\n')
	out.write ('# Column "Ign" denotes number of ignored reads due to indels (usually bad mapping!)\n')
	
	if rmax > 0:
		out.write ('# Max number of reads considered at a coordinate maxr= '+ str(maxr)+'. This limit is added to column "Ign", if the limit is reached at a coordinate.\n')
	
	out.write ('# Column "RefSeqcontext" denotes the poi-Refbase and the +-6 Refbases\n')
	out.write ('# Column "Class" denotes context class (e.g. stretches or STRs)\n')
	out.write ('# Date: '+ outtime+'\n')
	
	c = '# \t\t\t\t\t\t' + \
	'\tMapQV>RQVmin and MM<=MMmax and Read>MLmin and BaseQVs>=QVmin\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t' + \
	'\tMM<=MMmax and Read>MLmin and BaseQVs>=QVmin\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t' + \
	'\tRead>MLmin and BaseQVs>=QVmin\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t' + \
	'\tBaseQVs>=QVmin\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t' + \
	'\tNo QV-Filtering\n'
	out.write (c)
	
	c = '# \t\t\t\t\t\t' + \
	'\tUnique start points (on +- strands)\t\t\t\t\t\t\t\t\t\tRemaining coverage\t\t\t\t\t\t\t\t\t' + \
	'\tUnique start points (on +- strands)\t\t\t\t\t\t\t\t\t\tRemaining coverage\t\t\t\t\t\t\t\t\t' + \
	'\tUnique start points (on +- strands)\t\t\t\t\t\t\t\t\t\tRemaining coverage\t\t\t\t\t\t\t\t\t' + \
	'\tUnique start points (on +- strands)\t\t\t\t\t\t\t\t\t\tRemaining coverage\t\t\t\t\t\t\t\t\t' + \
	'\tUnique start points (on +- strands)\t\t\t\t\t\t\t\t\t\tCoverage\t\t\t\t\t\t\t\t\t\n'
	out.write (c)
	
	c = '# C\tPos[1]\tRef\tRefSeqContext\tGC[%]\tClass\tIgn' + \
		'\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-'+ \
		'\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-'+ \
		'\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-'+ \
		'\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-'+ \
		'\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\n'
	out.write (c)

	#######################################################################################################
	# Start of WHILE loop over all positions of interest
	#######################################################################################################

	while True:

		#######################################################################################
		#  read short fasta sequence [poi-LFR; poi +LFR] from refseq file
		#######################################################################################
	
		refseq=open(refseqfile, 'rb')
		# testing:
		# refseq.seek(6)
		# reads 7th character in file:
		# t=refseq.read(1)
		# reads 7th-8th characters in file:
		# t=refseq.read(2)
		# print t
		# sys.exit()
	
		# max aligned read length (including short INDELS)
		# LFR=50+100
		# but in this version we ignore reads with indels because many
		# current gapped alignment methods create more alignment artifacts
		# than sensible reads:
		# LFR=50
	
		# chrom length and start in refseq fasta file
		lmax,k = fai[chrom]
		# line length (excl. EOL) in fasta file (must be constant):
		refseq.seek(k)
		t=refseq.read(1000)
		ll=t.find('\n')
		# mapped read start and end in chromosome:
		j1=max(0,poi+1-LFR)
		j2=min(poi+LFR,lmax)
		# mapped read start and end in refseq fasta file:
		nn=int(j1/ll)
		rr= j1 % ll
		jj1=k+nn*(ll+1)+rr
		nn=int(j2/ll)
		rr= j2 % ll
		jj2=k+nn*(ll+1)+rr
		# number of bytes
		n=jj2-jj1+1
		# print jj1, jj2, n
		# read n bytes from position jj1, discarding the EOLs:
		# (as strip does not work for \n)
		# t=refseq.read(n)
		# t=t.strip('\n')
		# print t
		refseq.seek(jj1)
		t=''
		for jj in range(0,n):
			tt=refseq.read(1)
			if tt != '\n':
				t+=tt
	
		#######################################################################################
		# Get reference base at poi:
		#######################################################################################
	
		#print 't='
		#print t
		kpoi=poi-j1
		#print ' poi=%s, kpoi =%s ' % (poi, kpoi)
		refbasepoi = t[kpoi:kpoi+1]
		#print ' refbasepoi = %s' % (refbasepoi)
		# sys.exit()
	
		#######################################################################################
		# Get context, GC-content and sequence classification at poi:
		#######################################################################################

		# 2011_02_04
		lt=len(t)-1
		ctxt1= t[max(0,kpoi-6):max(0,kpoi)]
		ctxt2 = t[min(kpoi+1,lt):min(kpoi+7,lt)]
		ctxt = ctxt1+'['+refbasepoi+']' + ctxt2
		#print ctxt
		#print t[min(kpoi+1,lt):min(kpoi+5,lt)]
		#print kpoi, LFR, min(kpoi+1,lt), min(kpoi+5,lt)
		# ctxt += t[min(kpoi+1,lt):min(kpoi+7,lt)]
		#print ctxt

		# GC-content in %:
		gcc = 100*(ctxt.count('G')+ctxt.count('C')+ctxt.count('g')+ctxt.count('c'))
		try:
			gcc=gcc/(len(ctxt)-2)
		except:
			print '*** Exception in line 603: gcc=gcc/(len(ctxt)-2)'
			print '*** fasta file name = %s' % (refseqfile)
			print 'chrom=%s, poi=%s, ctxt=%s, len(ctxt)=%s, kpoi=%s, refbasepoi=%s, lt=%s' % (chrom, poi+1, ctxt, len(ctxt), kpoi, refbasepoi, lt)

		#### Sequence classification: ####

		# convert string to upper case
		# 2011_02_04
		# ctxtu=ctxt.upper()
		ctxtu = ctxt1 + refbasepoi + ctxt2
		ctxtu = ctxtu.upper()
		
		cla=0

		# find single nucleotide repeats:
		k=-1
		for poicl in poicls:
			k+=1
			if ctxtu.find(poicl) > -1 :
				cla=k

		# find short tandem repeats:
		clat=0
		k=-1
		for poicl in poiclt:
			k+=1
			if ctxtu.find(poicl) > -1 :
				clat=k

		# Merge single and tandem repeats:
		cla = cla + 100*clat

		if cla == 0:
			# no single-nucleotide or tandem-nucleotide repeats
			cla=1

		#######################################################################################
		# experiments from http://wwwfgu.anat.ox.ac.uk/~andreas/documentation/samtools/api.html#module-pysam
		# on small file strategy_abdou_755bc14_saet_F3.sorted.bam
		#######################################################################################
	
		# contig/chrom-name required. start and end give the read-starts (0-based) and must be a range
		#for read in samfile.fetch('BRAC1_BRAC2_refseq',start=50, end=60):
			# print read.pos
		#	print read
	
		# remember read names which are already processed:
		reads=[]
		reads.append('dummy')
	
		# max fragment length (including short INDELS)
		# LFR=50+100
		# but in a first version we ignore reads with indels because the
		# the current indel detection tools appear to create more artefacts than
		# sensible reads:
		# LFR=50
	
		# This following code block is a bit quick and dirty, but:
		#   + much easier to understand,
		#   + much easier to write,
		#   + much easier to test and debug,
		#   + faster to run
		#   than a 3-dimensional array with nested loops
	
		# count unique start points - re-initialise lists
		# for ACGTN, for strands 1 and 2, and for
		#    - all bases and
		#    - QV-limited bases and
		#    - QV-limited bases and mapped-length-of-read threshold and
		#    - QV-limited bases, mapped-length-of-read, and mismatch-in-read threshold
		#    - QV-limited bases, mapped-length-of-read, mismatch-in-read, and read-mapping QV threshold
		
		for j in range(0,LFR):
			isa1[j]=0
			isa2[j]=0
			isqa1[j]=0
			isqa2[j]=0
			isma1[j]=0
			isma2[j]=0
			ismma1[j]=0
			ismma2[j]=0
			ismqa1[j]=0
			ismqa2[j]=0
	
			isc1[j]=0
			isc2[j]=0
			isqc1[j]=0
			isqc2[j]=0
			ismc1[j]=0
			ismc2[j]=0
			ismmc1[j]=0
			ismmc2[j]=0
			ismqc1[j]=0
			ismqc2[j]=0
	
			isg1[j]=0
			isg2[j]=0
			isqg1[j]=0
			isqg2[j]=0
			ismg1[j]=0
			ismg2[j]=0
			ismmg1[j]=0
			ismmg2[j]=0
			ismqg1[j]=0
			ismqg2[j]=0
	
			ist1[j]=0
			ist2[j]=0
			isqt1[j]=0
			isqt2[j]=0
			ismt1[j]=0
			ismt2[j]=0
			ismmt1[j]=0
			ismmt2[j]=0
			ismqt1[j]=0
			ismqt2[j]=0
	
			isn1[j]=0
			isn2[j]=0
			isqn1[j]=0
			isqn2[j]=0
			ismn1[j]=0
			ismn2[j]=0
			ismmn1[j]=0
			ismmn2[j]=0
			ismqn1[j]=0
			ismqn2[j]=0
	
		# coverage count at poi for ACGTN, for strands 1 and 2 (without QV-limit):
		ica1=0
		ica2=0
	
		icc1=0
		icc2=0
	
		icg1=0
		icg2=0
	
		ict1=0
		ict2=0
	
		icn1=0
		icn2=0
	
		# coverage count at poi for ACGTN, for strands 1 and 2 (with QV-limit):
		icqa1=0
		icqa2=0
	
		icqc1=0
		icqc2=0
	
		icqg1=0
		icqg2=0
	
		icqt1=0
		icqt2=0
	
		icqn1=0
		icqn2=0
	
		# coverage count at poi for ACGTN, for strands 1 and 2 (with QV-limit and ML-limit):
		icma1=0
		icma2=0
	
		icmc1=0
		icmc2=0
	
		icmg1=0
		icmg2=0
	
		icmt1=0
		icmt2=0
	
		icmn1=0
		icmn2=0
	
		# coverage count at poi for ACGTN, for strands 1 and 2 (with QV-limit, ML-limit, MM-Limit):
		icmma1=0
		icmma2=0
	
		icmmc1=0
		icmmc2=0
	
		icmmg1=0
		icmmg2=0
	
		icmmt1=0
		icmmt2=0
	
		icmmn1=0
		icmmn2=0
	
	
		# coverage count at poi for ACGTN, for strands 1 and 2 (with QV-limit, ML-limit, MM-Limit,
		# and read-mapping QV-Limit):
		icmqa1=0
		icmqa2=0
	
		icmqc1=0
		icmqc2=0
	
		icmqg1=0
		icmqg2=0
	
		icmqt1=0
		icmqt2=0
	
		icmqn1=0
		icmqn2=0
	
	
		# user-specified base quality value threshold
		# below qvmin for the base at the poi, the read's start point is not used in isq
		# and the read is not used for coverage at the poi in icq
		# qvmin=int(sys.argv[3])
	
		# user-specified mapped-length-of-read threshold (in addition to qvmin threshold)
		# below MLmin, the read's start point is not used in ism
		# and the read is not used for coverage at the poi in icm
		# MLmin=int(sys.argv[4])
	
	
		# for pileupcolumn in samfile.pileup('BRAC1_BRAC2_refseq',start=118780, end=118781):
		# for pileupcolumn in samfile.pileup('BRAC1_BRAC2_refseq',50,51):
	
		#######################################################################################
		#
		#    LOOP OVER ALL READS IN SAM PILEUP-COLUMN AT POSITION OF INTEREST
		#
		#######################################################################################
	
		# number of reads which are longer than LFR:
		toolong=0
	
		# number of reads with indel(s):
		indelread=0
	
		# Current 1-based chromosomal coord for which read-startpoints (SAM-startpoints) are counted
		k=0
		for pileupcolumn in samfile.pileup(chrom, poi, poi+1):
			for pileupread in pileupcolumn.pileups:
				
				# sam read flag (required, because BWA truncates read names, i.e. 
				# not possible to distinguish ends of paired-end reads by sam file name):
				flg = int(pileupread.alignment.flag)
				
				# check for unmapped-read bit:
				if flg&4 == 4:
					# this read is unmapped, and in BAM-files from BWA the remaining read info is junk!!
					continue
				
				# fragment-of-template info (Illumina-BWA paired-ends): 
				# test bits 6 and 7 of sam flag
				if flg&64 == 64:
					frag=':1'
				elif flg&128 == 128:
					frag=':2'
				else:
					frag=''		
				# name of read:
				name = pileupread.alignment.qname + frag
				
		
				# Bioscope 1.0.1 and Bioscope 1.2.1 CIGARS do not contain mismatch info,
				# SO USE MD-TAG-CIGARs

				# 18 October 2010 - except in some MP-cases??
		
				cigarp = pileupread.alignment.cigar
				# print cigarp  -->  [(0, 47), (5, 3)] --> parse first number of tuples in list for indels
				#if name == 'HWI-ST751_0067:4:2204:2424:166228#GCCAAT:2':
				#	print cigarp
				# --> [(4, 22), (0, 79)] == 22S79M  (softclipping at start of read; pysam bug: hardclips the qvals)
				# sys.exit()
				
				iskip=0
				for indeltype, number in cigarp:
					if indeltype == 1 or indeltype == 2:
						# 1=I, 2=D
						iskip=1
						#print ' read: %s, indeltype: %s' % (name, indeltype)
						break

				if iskip == 0:
					# get MD-tag cigar:
					cigar = pileupread.alignment.opt('MD')
					if cigar.find('I') > -1 or cigar.find('D') > -1 or cigar.find('^') > -1:
						#print ' read: %s, cigarIndel: found I or D or esc-symbol' % (name)
						iskip = 1
		
				if iskip == 1:
					# ignore this read completely , but remember it and count it
					#print ' --------- read skipped -------------------------'
					
					if not name in reads:
						# get read information only once, because same read occurs up to LFR times in pileup
						reads.append(name)
						if maxr > 0 and len(reads) > maxr:
							indelread += maxr
						# count this read
						indelread += 1
				elif not name in reads:
				# if not name in reads:
					# get read information only once, because same read occurs up to LFR times in pileup
					reads.append(name)
					if maxr > 0 and len(reads) > maxr:
						indelread += maxr
	
					# count base space mismatches in read:
					mm = cigar.count('A') + cigar.count('C') + cigar.count('G') + cigar.count('T') + \
					     cigar.count('N') + cigar.count('a') + cigar.count('c') + cigar.count('g') + \
					     cigar.count('t') + cigar.count('n')
	
					# fasta sequence, truncated
					seq = pileupread.alignment.seq
	
					# start point of truncated fasta sequence (0-based!)
					startpt = pileupread.alignment.pos
	
					# position of interest (poi) in local coords of read (0-based):
					qp = pileupread.qpos
					poipos = poi - startpt + qp
					# 20 October 2010  Due to following problem
					# [bam_plp_destroy] memory leak: 2. Continue anyway.
					# Traceback (most recent call last):
					# File "/work_F0/sukmb222/software/bin/pibase", line 1483, in <module>
					# pib (details, verbose, onechr, onepoi, refseqfile, tablefile, bamfile, outfile, LFR, QVmin, MLmin, MMmax, RQVmin)
					# File "/work_F0/sukmb222/software/bin/pibase", line 849, in pib
					# base = seq[poipos]
					# IndexError: string index out of range
					# ### skip such problem reads ####
					if poipos < 0 or poipos >= len(seq):
						continue
	
					# base call of this read at the poi
					base = seq[poipos]
	
					# base quality value for truncated fasta sequence
					qvals = pileupread.alignment.qqual
					
					# work-around for pysam soft-clipping bug (qvals is hard-clipped):
					qpoipos = poipos
					if len(seq) != len(qvals):
						itype, icount = cigarp[0]
						#if name == 'HWI-ST751_0067:4:2204:2424:166228#GCCAAT:2':
						#	print name, itype, icount
						qpoipos = poipos - icount
	
					# quality value for base at the poi
					try:
						#qv = ord(qvals[poipos]) - 33
						qv = ord(qvals[qpoipos]) - 33
					#except:
					except IndexError:
						# we get invalid qvals at some positions in the 1000Genomes Illumina and FLX BAM files
						# so: ignore this read completely but count as Ignored:
						print '-------------- exception: qvals[qpoipos] -------------'
						print ' name: %s, poipos: %s, qpoipos: %s' % (name, poipos, qpoipos)
						print ' len(seq): %s,  len(qvals):  %s' % (len(seq), len(qvals)) 
						print ' seq  : %s' % (seq) 
						print ' qvals: %s' % (qvals)
						indelread += 1
						continue
	
					# mapping quality value for read
					rqv = pileupread.alignment.mapq
	
					# Length of the aligned query sequence
					qlen = pileupread.alignment.qlen
					if qlen > LFR:
						# ignore this read because it is longer than LFR:
						# so: ignore this read completely but count as Ignored:
						# print ' ----------- read longer than LFR -----------------'
						indelread += 1
						toolong += 1
						continue
	
					# sam read flag
					#flg = int(pileupread.alignment.flag)
	
					# Strand info: test bit 4 of flag
					if flg&16 == 16:
						strand=2
					else:
						strand=1
	
					# This following code block is a bit quick and dirty, but:
					#   + much easier to understand,
					#   + much easier to write,
					#   + much easier to test and debug,
					#   + faster to run
					#   than a 3-dimensional array with nested loops
	
					# Strand-specific and QV-specific unique start points and coverage counts:
					if base == 'A':
						if strand == 1:
							j = poi - startpt
							isa1[j] = 1
							ica1 += 1
							if qv >= qvmin:
								isqa1[j] = 1
								icqa1 += 1
								if qlen >= MLmin:
									# print ' yes'
									isma1[j] = 1
									icma1 += 1
									if mm <= MMmax:
										ismma1[j] = 1
										icmma1 += 1
										if rqv >= RQVmin:
											ismqa1[j] = 1
											icmqa1 += 1
						else:
							endpt = startpt + qlen - 1
							j = endpt - poi
							# if name=='322_33_86':
							# print name+' j=%s endpt=%s startpt=%s qlen=%s' % (j, endpt, startpt, qlen)
							isa2[j] = 1
							ica2 += 1
							if qv >= qvmin:
								isqa2[j] = 1
								icqa2 += 1
								if qlen >= MLmin:
									isma2[j] = 1
									icma2 += 1
									if mm <= MMmax:
										ismma2[j] = 1
										icmma2 += 1
										if rqv >= RQVmin:
											ismqa2[j] = 1
											icmqa2 += 1
					elif base == 'C':
						if strand == 1:
							j = poi - startpt
							isc1[j] = 1
							icc1 += 1
							if qv >= qvmin:
								isqc1[j] = 1
								icqc1 += 1
								if qlen >= MLmin:
									ismc1[j] = 1
									icmc1 += 1
									if mm <= MMmax:
										ismmc1[j] = 1
										icmmc1 += 1
										if rqv >= RQVmin:
											ismqc1[j] = 1
											icmqc1 += 1
						else:
							endpt = startpt + qlen
							j = endpt - poi
							isc2[j] = 1
							icc2 += 1
							if qv >= qvmin:
								isqc2[j] = 1
								icqc2 += 1
								if qlen >= MLmin:
									ismc2[j] = 1
									icmc2 += 1
									if mm <= MMmax:
										ismmc2[j] = 1
										icmmc2 += 1
										if rqv >= RQVmin:
											ismqc2[j] = 1
											icmqc2 += 1
					elif base == 'G':
						if strand == 1:
							j = poi - startpt
							isg1[j] = 1
							icg1 += 1
							if qv >= qvmin:
								isqg1[j] = 1
								icqg1 += 1
								if qlen >= MLmin:
									ismg1[j] = 1
									icmg1 += 1
									if mm <= MMmax:
										ismmg1[j] = 1
										icmmg1 += 1
										if rqv >= RQVmin:
											ismqg1[j] = 1
											icmqg1 += 1
						else:
							endpt = startpt + qlen
							j = endpt - poi
							isg2[j] = 1
							icg2 += 1
							if qv >= qvmin:
								isqg2[j] = 1
								icqg2 += 1
								if qlen >= MLmin:
									ismg2[j] = 1
									icmg2 += 1
									if mm <= MMmax:
										ismmg2[j] = 1
										icmmg2 += 1
										if rqv >= RQVmin:
											ismqg2[j] = 1
											icmqg2 += 1
					elif base == 'T':
						if strand == 1:
							j = poi - startpt
							ist1[j] = 1
							ict1 += 1
							if qv >= qvmin:
								isqt1[j] = 1
								icqt1 += 1
								if qlen >= MLmin:
									ismt1[j] = 1
									icmt1 += 1
									if mm <= MMmax:
										ismmt1[j] = 1
										icmmt1 += 1
										if rqv >= RQVmin:
											ismqt1[j] = 1
											icmqt1 += 1
						else:
							endpt = startpt + qlen
							j = endpt - poi
							ist2[j] = 1
							ict2 += 1
							if qv >= qvmin:
								isqt2[j] = 1
								icqt2 += 1
								if qlen >= MLmin:
									ismt2[j] = 1
									icmt2 += 1
									if mm <= MMmax:
										ismmt2[j] = 1
										icmmt2 += 1
										if rqv >= RQVmin:
											ismqt2[j] = 1
											icmqt2 += 1
					elif base == 'N':
						if strand == 1:
							j = poi - startpt
							isn1[j] = 1
							icn1 += 1
							######################################
							# does N get a QV at all??
							# if not, then if-block is wrong
							######################################
							if qv >= qvmin:
								isqn1[j] = 1
								icqn1 += 1
								if qlen >= MLmin:
									ismn1[j] = 1
									icmn1 += 1
									if mm <= MMmax:
										ismmn1[j] = 1
										icmmn1 += 1
										if rqv >= RQVmin:
											ismqn1[j] = 1
											icmqn1 += 1
						else:
							endpt = startpt + qlen
							j = endpt - poi
							isn2[j] = 1
							icn2 += 1
							######################################
							# does N get a QV at all??
							# if not, then if-block is wrong
							######################################
							if qv >= qvmin:
								isqn2[j] = 1
								icqn2 += 1
								if qlen >= MLmin:
									ismn2[j] = 1
									icmn2 += 1
									if mm <= MMmax:
										ismmn1[j] = 1
										icmmn1 += 1
										if rqv >= RQVmin:
											ismqn2[j] = 1
											icmqn2 += 1
	
					# Alternatively: is_reverse means that read is mapped to strand 2:
					# reverse_strand = pileupread.alignment.is_reverse
					# strand=1
					# if reverse_strand:
					#	strand=2
	
	
					# update start point lists
	
					if startpt > k:
						k=startpt
						if verbose and details:
							print '################## start of pos.'+str(k+1)+'####################'
					if verbose and details:
						print '\tbase@%s of chr# %s in read %s=%s QV=%s readlen=%s, strand=%s, MDCIGAR=%s, MM=%s MapQV=%s' % (poi+1, ichrom+1, name, seq[poipos], qv, qlen, strand, cigar, mm, rqv)
	
				# print pileupread.alignment.pos, pileupread.qpos
				# print '\tbase in read %s = %s   base quality in read = %' % (pileupread.alignment.qname, pileupread.alignment.seq[pileupread.qpos], pileupread.alignment.qqual[pileupread.qpos])
				# print '\tbase in read %s = %s ' % (pileupread.alignment.qname, pileupread.alignment.seq[pileupread.qpos])
				# print '\tbase quality =' +str(pileupread.alignment.qqual)
				
				if maxr > 0 and indelread >= maxr:
					# print ' break out of loop'
					break
			if maxr > 0 and indelread >= maxr:
				# print ' break out of loop'
				# needs two breaks because this pysam iterator does odd things when coverages are very high!!
				break
	
		# This following code block is a bit quick and dirty, but:
		#   + much easier to understand,
		#   + much easier to write,
		#   + much easier to test and debug,
		#   + faster to run
		#   than a 3-dimensional array with nested loops
	
		# sum unique start points for the poi (strand 1 and 2, without and with qual value threshold):
		issa1=0
		issa2=0
		issqa1=0
		issqa2=0
		issma1=0
		issma2=0
		issmma1=0
		issmma2=0
		issmqa1=0
		issmqa2=0
	
		issc1=0
		issc2=0
		issqc1=0
		issqc2=0
		issmc1=0
		issmc2=0
		issmmc1=0
		issmmc2=0
		issmqc1=0
		issmqc2=0
	
		issg1=0
		issg2=0
		issqg1=0
		issqg2=0
		issmg1=0
		issmg2=0
		issmmg1=0
		issmmg2=0
		issmqg1=0
		issmqg2=0
	
		isst1=0
		isst2=0
		issqt1=0
		issqt2=0
		issmt1=0
		issmt2=0
		issmmt1=0
		issmmt2=0
		issmqt1=0
		issmqt2=0
	
		issn1=0
		issn2=0
		issqn1=0
		issqn2=0
		issmn1=0
		issmn2=0
		issmmn1=0
		issmmn2=0
		issmqn1=0
		issmqn2=0
		for j in range(0,LFR):
			issa1 += isa1[j]
			issa2 += isa2[j]
			issqa1 += isqa1[j]
			issqa2 += isqa2[j]
			issma1 += isma1[j]
			issma2 += isma2[j]
			issmma1 += ismma1[j]
			issmma2 += ismma2[j]
			issmqa1 += ismqa1[j]
			issmqa2 += ismqa2[j]
	
			issc1 += isc1[j]
			issc2 += isc2[j]
			issqc1 += isqc1[j]
			issqc2 += isqc2[j]
			issmc1 += ismc1[j]
			issmc2 += ismc2[j]
			issmmc1 += ismmc1[j]
			issmmc2 += ismmc2[j]
			issmqc1 += ismqc1[j]
			issmqc2 += ismqc2[j]
	
			issg1 += isg1[j]
			issg2 += isg2[j]
			issqg1 += isqg1[j]
			issqg2 += isqg2[j]
			issmg1 += ismg1[j]
			issmg2 += ismg2[j]
			issmmg1 += ismmg1[j]
			issmmg2 += ismmg2[j]
			issmqg1 += ismqg1[j]
			issmqg2 += ismqg2[j]
	
			isst1 += ist1[j]
			isst2 += ist2[j]
			issqt1 += isqt1[j]
			issqt2 += isqt2[j]
			issmt1 += ismt1[j]
			issmt2 += ismt2[j]
			issmmt1 += ismmt1[j]
			issmmt2 += ismmt2[j]
			issmqt1 += ismqt1[j]
			issmqt2 += ismqt2[j]
	
			issn1 += isn1[j]
			issn2 += isn2[j]
			issqn1 += isqn1[j]
			issqn2 += isqn2[j]
			issmn1 += ismn1[j]
			issmn2 += ismn2[j]
			issmmn1 += ismmn1[j]
			issmmn2 += ismmn2[j]
			issmqn1 += ismqn1[j]
			issmqn2 += ismqn2[j]
	
		if verbose:
			print '\n'
			print '######################################################################################'
			print ' POSITION INFO (UNIQUE START POINTS & COVERAGE PER STRAND)'
			print ' AT 1-BASED POSITION %s ON CHROM/CONTIG %s - "%s"' % (poi+1, ichrom+1, chrom)
			print ' REFERENCE: %s' % (refseqfile)
			print ' INPUT-FILE: %s' % (bamfile)
			print ' OUTPUT-FILE: %s' % (outfile)
			print ' MAX FRAGMENT LENGTH CONSIDERED: %s' % (LFR)
			print ' IGNORED READS BECAUSE WITH INDELS: %s' % (indelread)
			print ' IGNORED READS BECAUSE TOO LONG: %s' % (toolong)
			#print ' REFERENCE BASE: %s \t CONTEXT: %s \t GC:%spc \tClass:%s' % (refbasepoi, ctxt, gcc, cla)
			print ' REFERENCE BASE: %s \tCONTEXT: %s \tGC-CONTENT:%s%% \tCLASS:%s' % (refbasepoi, ctxt, gcc, cla)
			print ' '
			print ' UNIQUE START POINTS ON +STRAND AND -STRAND (MEASURE FOR INDEPENDENT FRAGMENTS):'
			print ' + QVMIN=%s,MLmin=%s,MMmax=%s,RQVmin=%s: \tA:%s \tC:%s \tG:%s \tT:%s \tN:%s' % (qvmin, MLmin, MMmax, RQVmin, issmqa1, issmqc1, issmqg1, issmqt1, issmqn1)
			print ' - QVMIN=%s,MLmin=%s,MMmax=%s,RQVmin=%s: \tA:%s \tC:%s \tG:%s \tT:%s \tN:%s' % (qvmin, MLmin, MMmax, RQVmin, issmqa2, issmqc2, issmqg2, issmqt2, issmqn2)
			print ' + QVMIN=%s,MLmin=%s,MMmax=%s: \t\t\tA:%s \tC:%s \tG:%s \tT:%s \tN:%s' % (qvmin, MLmin, MMmax, issmma1, issmmc1, issmmg1, issmmt1, issmmn1)
			print ' - QVMIN=%s,MLmin=%s,MMmax=%s: \t\t\tA:%s \tC:%s \tG:%s \tT:%s \tN:%s' % (qvmin, MLmin, MMmax, issmma2, issmmc2, issmmg2, issmmt2, issmmn2)
			print ' + QVMIN=%s,MLmin=%s: \t\t\t\tA:%s \tC:%s \tG:%s \tT:%s \tN:%s' % (qvmin, MLmin, issma1, issmc1, issmg1, issmt1, issmn1)
			print ' - QVMIN=%s,MLmin=%s: \t\t\t\tA:%s \tC:%s \tG:%s \tT:%s \tN:%s' % (qvmin, MLmin, issma2, issmc2, issmg2, issmt2, issmn2)
			print ' + QVMIN=%s: \t\t\t\t\tA:%s \tC:%s \tG:%s \tT:%s \tN:%s' % (qvmin, issqa1, issqc1, issqg1, issqt1, issqn1)
			print ' - QVMIN=%s: \t\t\t\t\tA:%s \tC:%s \tG:%s \tT:%s \tN:%s' % (qvmin, issqa2, issqc2, issqg2, issqt2, issqn2)
			print ' + QVMIN=0: \t\t\t\t\tA:%s \tC:%s \tG:%s \tT:%s \tN:%s' % (issa1, issc1, issg1, isst1, issn1)
			print ' - QVMIN=0: \t\t\t\t\tA:%s \tC:%s \tG:%s \tT:%s \tN:%s' % (issa2, issc2, issg2, isst2, issn2)
			print ' '
			print ' COVERAGE ON +STRAND AND -STRAND (MEASURE FOR SEQUENCING / TECHNICAL REPRODUCIBILITY):'
			print ' + QVMIN=%s,MLmin=%s,MMmax=%s,RQVmin=%s: \tA:%s \tC:%s \tG:%s \tT:%s \tN:%s' % (qvmin, MLmin,  MMmax, RQVmin, icmqa1, icmqc1, icmqg1, icmqt1, icmqn1)
			print ' - QVMIN=%s,MLmin=%s,MMmax=%s,RQVmin=%s: \tA:%s \tC:%s \tG:%s \tT:%s \tN:%s' % (qvmin, MLmin, MMmax, RQVmin, icmqa2, icmqc2, icmqg2, icmqt2, icmqn2)
			print ' + QVMIN=%s,MLmin=%s,MMmax=%s: \t\t\tA:%s \tC:%s \tG:%s \tT:%s \tN:%s' % (qvmin, MLmin,  MMmax, icmma1, icmmc1, icmmg1, icmmt1, icmmn1)
			print ' - QVMIN=%s,MLmin=%s,MMmax=%s: \t\t\tA:%s \tC:%s \tG:%s \tT:%s \tN:%s' % (qvmin, MLmin, MMmax, icmma2, icmmc2, icmmg2, icmmt2, icmmn2)
			print ' + QVMIN=%s,MLmin=%s: \t\t\t\tA:%s \tC:%s \tG:%s \tT:%s \tN:%s' % (qvmin, MLmin,  icma1, icmc1, icmg1, icmt1, icmn1)
			print ' - QVMIN=%s,MLmin=%s: \t\t\t\tA:%s \tC:%s \tG:%s \tT:%s \tN:%s' % (qvmin, MLmin, icma2, icmc2, icmg2, icmt2, icmn2)
			print ' + QVMIN=%s: \t\t\t\t\tA:%s \tC:%s \tG:%s \tT:%s \tN:%s' % (qvmin, icqa1, icqc1, icqg1, icqt1, icqn1)
			print ' - QVMIN=%s: \t\t\t\t\tA:%s \tC:%s \tG:%s \tT:%s \tN:%s' % (qvmin, icqa2, icqc2, icqg2, icqt2, icqn2)
			print ' + QVMIN=0: \t\t\t\t\tA:%s \tC:%s \tG:%s \tT:%s \tN:%s' % (ica1, icc1, icg1, ict1, icn1)
			print ' - QVMIN=0: \t\t\t\t\tA:%s \tC:%s \tG:%s \tT:%s \tN:%s' % (ica2, icc2, icg2, ict2, icn2)
			print '######################################################################################'


		#######################################################################################
		# write to tab-separated output file:
		#######################################################################################

		#c='%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t\n' % \
		#(ichrom+1, poi+1, refbasepoi, ctxt, gcc, cla, indelread, issqa1,issqa2,issqc1,issqc2,issqg1,issqg2,issqt1,issqt2,issqn1,issqn2, icqa1,icqa2,icqc1,icqc2,icqg1,icqg2,icqt1,icqt2,icqn1,icqn2, issa1,issa2,issc1,issc2,issg1,issg2,isst1,isst2,issn1,issn2, ica1,ica2,icc1,icc2,icg1,icg2,ict1,ict2,icn1,icn2)

		c='%s\t%s\t%s\t%s\t%s\t%s\t%s' % (ichrom+1, poi+1, refbasepoi, ctxt, gcc, cla, indelread)

		# For RQVmin, MMmax, MLmin, QVmin thresholds : unique start points:
		c+='\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s' % \
		(issmqa1, issmqa2,  issmqc1, issmqc2,  issmqg1, issmqg2,  issmqt1, issmqt2,  issmqn1, issmqn2)

		# For RQVmin, MMmax, MLmin, QVmin thresholds : coverage:
		c+='\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s' % \
		(icmqa1, icmqa2,  icmqc1, icmqc2,  icmqg1, icmqg2,  icmqt1, icmqt2,  icmqn1, icmqn2)

		# For MMmax, MLmin, QVmin thresholds : unique start points:
		c+='\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s' % \
		(issmma1, issmma2,  issmmc1, issmmc2,  issmmg1, issmmg2,  issmmt1, issmmt2,  issmmn1, issmmn2)
		
		# For MMmax, MLmin, QVmin thresholds : coverage:
		c+='\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s' % \
		(icmma1, icmma2,  icmmc1, icmmc2,  icmmg1, icmmg2,  icmmt1, icmmt2,  icmmn1, icmmn2)
		
		# For MLmin, QVmin thresholds : unique start points:
		c+='\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s' % \
		(issma1, issma2,  issmc1, issmc2,  issmg1, issmg2,  issmt1, issmt2,  issmn1, issmn2)

		# For MLmin, QVmin thresholds : coverage:
		c+='\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s' % \
		(icma1, icma2,  icmc1, icmc2,  icmg1, icmg2,  icmt1, icmt2,  icmn1, icmn2)

		# For QVmin threshold : unique start points:
		c+='\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s' % \
		(issqa1, issqa2,  issqc1, issqc2,  issqg1, issqg2,  issqt1, issqt2,  issqn1, issqn2)

		# For QVmin threshold : coverage:
		c+='\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s' % \
		(icqa1, icqa2,  icqc1, icqc2,  icqg1, icqg2,  icqt1, icqt2,  icqn1, icqn2)

		# For no filtering : unique start points:
		c+='\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s' % \
		(issa1, issa2,  issc1, issc2,  issg1, issg2,  isst1, isst2,  issn1, issn2)

		# For no filtering : coverage:
		c+='\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s' % \
		(ica1, ica2,  icc1, icc2,  icg1, icg2,  ict1, ict2,  icn1, icn2)

		# End of line:
		c+='\n'
		out.write (c)

		#######################################################################################
		# Get next position of interest from table file
		#######################################################################################

		if onepoi > 0:
			# print ' break out of loop'
			break
		else:
	
			# print ' step 1'
			eof=False
			while True:
				line=tabf.readline()
				line=line.strip()
				# print line+'\n'
				# print len(line)
				if not line.startswith('#'):
					if len(line)==0:
						#print 'EOF'
						# end of file
						eof=True
					break
			if eof:
				#print 'EOF'
				break
	
			if tabfiletype==1:
				# Get ichrom (0-based) and poi (0-based) from Bioscope diBayes gff3 file:
				try:
					cols = line.split('\t')
					chromn = cols[0][3:]
					ichrom = int(chromn.strip()) - 1
					poi = int(cols[3].strip()) - 1
				except:
					sys.exit('*** Error: Bad diBayes gff3 format in SNP-line of file:  %s\n***        Line is:\n%s\n' % (tablefile,line))
			elif tabfiletype==3:
				# Get ichrom (0-based) and poi (0-based) from samtools pileup file:
				try:
					cols = line.split('\t')
					# Pre-v1.4.2:
					# chromn = cols[0][3:]
					# ichrom = int(chromn.strip()) - 1
					# v1.4.2: Find chromosome name in bam-file header:
					chromn = cols[0].strip()
					if chromn in refs:
						ichrom = refs.index(chromn)
					else:
						print '*** Error: chromosome/contig %s is unknown in BAM file.\n***        Line is:\n%s\n' % (chromn,line)
						sys.exit()
					poi = int(cols[1].strip()) - 1
				except:
					if vcfformat:
						sys.exit('*** Error: Bad vcf format in SNP-line of file:  %s\n***        Line is:\n%s\n' % (tablefile,line))
					else:
						sys.exit('*** Error: Bad samtools pileup format in SNP-line of file:  %s\n***        Line is:\n%s\n' % (tablefile,line))
			else:
				# Get ichrom (0-based) and poi (0-based) from tab-separated file:
				try:
					cols = line.split('\t')
					chromn = cols[0].strip()
					if chromn in refs:
						ichrom = refs.index(chromn)
					else:
						print '*** Error: chromosome/contig %s is unknown in BAM file.\n***        Line is:\n%s\n' % (chromn,line)
						sys.exit()
					poi = int(cols[1].strip()) - 1
				except:
					print '*** Error: Bad format in SNP-line of file:  %s\n***        Line is:\n%s\n' % (tablefile,line)
					sys.exit('***        Line should be: <chromosome_name> TAB <1_based_position_in_chrom>')
	
			# Do not allow jumping to previous chromosome (for speed reasons in a later version...):
			if ichrom < ichrommax:
				sys.exit('*** Error: Small chromosome number %s after large chromosome number %s not permitted.\nIn file:  %s\nLine:  %s\n' % (ichrom, ichrommax, tablefile, line))
			elif ichrom > ichrommax:
				ichrommax=ichrom
	
			# Translate chromosome number into chromosome name:
			if ichrom+1 > nrefs:
				if nrefs == 1:
					sys.exit('*** Error: Bad <chr#> = %s in file, because only 1 chrom/contig is defined in bamfile header.\nLine:\n%s\n' % (ichrom+1, line))
				else:
					sys.exit('*** Error: Bad <chr#> = %s in file, because only %s chroms/contigs are defined in bamfile header.\nLine:\n%s\n' % (ichrom+1, nrefs, line))
			chrom = refs[ichrom]
	
	#######################################################################################################
	# End of WHILE loop over all positions of interest
	#######################################################################################################

	out.close()

###########################################################################################################
#
# the actual main
#
###########################################################################################################

version = '1.4.7'

###########################################################################################################
# set resource limits:
###########################################################################################################

setlimits()

###########################################################################################################
# get command line arguments
###########################################################################################################

nargs = len(sys.argv)
if nargs < 6:
	print pib.__doc__
	sys.exit()

# defaults
verbose = False
details = False
onechr = 0
onepoi = 0
QVmin = 20
MLmin = 49
MMmax = 1
RQVmin = 20

# optional -vdp argument:
if sys.argv[1].startswith("-"):
	options = sys.argv[1]
	verbose = options.find('v') > -1
	details = options.find('d') > -1
	if options.find('p') > -1 :
		if nargs < 7:
			print pib.__doc__
			sys.exit()
		onechr = sys.argv[2]
		try:
			onepoi = int(sys.argv[3])
		except:
			print '\n*** Error: integer number required for <poi>\n'
			print pib.__doc__
			sys.exit()
		tablefile='dummy'
		k=4
	else:
		if nargs < 6:
			print pib.__doc__
			sys.exit()
		tablefile=sys.argv[2]
		k=3
else:
# no optional -vdp argument:
	tablefile=sys.argv[1]
	k=2

# required arguments:
refseqfile = sys.argv[k]
bamfile = sys.argv[k+1]
outfile = sys.argv[k+2]
LFR = max(int(sys.argv[k+3]),50)
maxr = -1

# optional trailing arguments:
if nargs > k+4:
	try:
		QVmin=int(sys.argv[k+4])
	except:
		print'\n*** Error: integer number required for <QVmin>\n'
		print pib.__doc__
		sys.exit()
if nargs > k+5:
	try:
		MLmin=int(sys.argv[k+5])
	except:
		print'\n*** Error: integer number required for <MLmin>\n'
		print pib.__doc__
		sys.exit()
if nargs > k+6:
	try:
		MMmax=int(sys.argv[k+6])
	except:
		print'\n*** Error: integer number required for <MMmax>\n'
		print pib.__doc__
		sys.exit()
if nargs > k+7:
	try:
		RQVmin=int(sys.argv[k+7])
	except:
		print'\n*** Error: integer number required for <RQVmin>\n'
		print pib.__doc__
		sys.exit()
if nargs > k+8:
	try:
		maxr=int(sys.argv[k+8])
	except:
		print'\n*** Error: integer number required for <maxr>\n'
		print pib.__doc__
		sys.exit()

#print (verbose)
#print (details)
#print (onepoi)

###########################################################################################################
# call main procedure
###########################################################################################################

pib (version, details, verbose, onechr, onepoi, refseqfile, tablefile, bamfile, outfile, LFR, QVmin, MLmin, MMmax, RQVmin, maxr)

print('*** pibase_bamref completed successfully')


