#!/usr/bin/python

import sys
import time
import resource

__author__="mforster"
__date__ ="$Mar 28, 2011 1:00 PM$"

def setlimits():
	#############################################################################
	# sets resource limits (mandatory for job interoperability on cluster)
	# to 10MB Stack + 1GB data + 1GB RLIMIT_RSS
	# first steps taken from
	# http://docs.python.org/library/resource.html and
	# http://www.doughellmann.com/PyMOTW/resource/
	# Author: MICHAEL FORSTER m.forster@ikmb.uni-kiel.de
	# Version: 13 September 2010
	#############################################################################

	#############################################################################
	# Adjust the maximum size (in bytes) of the process's heap
	#############################################################################

	soft, hard = resource.getrlimit(resource.RLIMIT_DATA)
	#print 'Data/heap limit starts as  :', soft

	# 1GB heap:
	newheap=1024*1024*1024

	resource.setrlimit(resource.RLIMIT_DATA, (newheap, hard))

	soft, hard = resource.getrlimit(resource.RLIMIT_DATA)
	#print 'Data/heap limit changed to :', soft

	#############################################################################
	# ca 10MB original stack changed to exactly 10MB stack:
	#############################################################################

	#newstack=10*1024*1024+1
	newstack=10*1024*1024
	soft, hard = resource.getrlimit(resource.RLIMIT_STACK)
	#print 'Stack limit starts as  :', soft
	resource.setrlimit(resource.RLIMIT_STACK, (newstack, hard))
	soft, hard = resource.getrlimit(resource.RLIMIT_STACK)
	#print 'Stack limit changed to :', soft

	#############################################################################
	# 1GB RLIMIT_RSS
	# "Maximale resident set size in Byte. Bei Speicherengpaessen entzieht der Kern
	# den Prozessen, die den resident set size ueberschreiten, den zuviel
	# angeforderten Speicher."
	# http://pronix.linuxdelta.de/C/Linuxprogrammierung/Linuxsystemprogrammieren_C_Kurs_Kapitel3a.shtml
	#############################################################################

	newrss=1024*1024*1024
	soft, hard = resource.getrlimit(resource.RLIMIT_RSS)
	#print 'RLIMIT_RSS starts as  :', soft
	resource.setrlimit(resource.RLIMIT_RSS, (newrss, hard))
	soft, hard = resource.getrlimit(resource.RLIMIT_RSS)
	#print 'RLIMIT_RSS changed to :', soft

def pib_c_to_contig (infile, infile2, outfile):
	'''        #######################################################################################
	#  pibase_c_to_contig v0.9
	#######################################################################################
	#  Convert pibase-contig-numbers into contig names (e.g. 25 -> chrM)
	#
	#  Usage:	pibase_c_to_contig <in> [in2] <out>
	#  where
	# 
	#  in:   file to be converted (from pibase_consensus, pibase_fisherdiff, ...)
	#        Note: pibase_fisherdiff has a shortened file head, so a second input file
	#        with the conversion table must be given.
	#
	#  in2:  file with contig numbering/name table in header
	#        (required, if the first input file is a pibase_fisherdiff file)
	#
	#  out:  output file 
	#######################################################################################
	#  Author: Michael Forster, m.forster@ikmb.uni-kiel.de         
	#  Date: 7 August 2012
	#######################################################################################
	'''
	
	print ' Infile:', infile 
	print ' Infile2:', infile2 
	print ' Outfile:', outfile
	
	############################################################################
	# open infile2 and find contig numbering/naming table in header 
	############################################################################
	
	inf = open (infile2, 'r')
	while True:
		line = inf.readline()
		if len(line) == 0:
			# EOF
			print ' *** Error: There is no table header line containing "# Mapping of pibase-chromosome-number to chromosome/contig name:"'
			sys.exit (' in file %s' % (infile2))
		if line.find('# Mapping of pibase-chromosome-number to chromosome/contig name:')> -1 :
			line = inf.readline()
			break
	
	############################################################################
	# read contig numbering/naming table into dictionary 
	############################################################################
	
	chrom = {}
	n = 0
	while True:
		line = inf.readline().strip()
		cols = line.split('\t')
		if len(cols) != 3 :
			break
		n += 1
		if int(cols[1]) != n :
			print ' *** Error: There is no continuous contig numbering'
			sys.exit (' in file %s' % (infile))
		chrom[n] = cols[2]
	if n == 0:
		print ' *** Error: There is no contig numbering/name table in the header lines'
		sys.exit (' in file %s' % (infile))
	
	inf.close()
	# print chrom
	
	############################################################################
	# open infile and test for file type 
	############################################################################
	
	ft = -1
	
	inf = open (infile, 'r')
	while True:
		line = inf.readline()
		if len(line) == 0:
			# EOF
			print ' *** Error: There is no table header line containing "RefSeqContext<TAB>GC[%]"'
			sys.exit (' in file %s' % (infile))
		if line.find('RefSeqContext\tGC[%]')> -1 :
			header = line
			break
	if header.startswith('# C\tPos[1]'):
		# pibase_consensus
		ft = 1
	elif header.startswith('#D\tD_error\tp.med'):
		# pibase_fisherdiff or pibase_fisherdiff_bestchoice
		if header.find('\tChr') > -1 :
			# pibase_fisherdiff_bestchoice
			ft = 2
		else:
			ft = 3
	elif header.startswith('#D\tC'):
		# pibase_diff
		ft = 3
	else:
		print ' *** Error: The table header line is inconsistent with a valid input file format.'
		print ' *** File: %s' % (infile)
		print ' *** Table header: ', header [:30]
		sys.exit (' *** Allowed file formats: pibase_consensus, pibase_fisherdiff, or pibase_diff output.')
	
	inf.close()

	############################################################################
	# find column chr
	############################################################################
	
	header = header.strip()
	# print header
	cols=header.split('\t')
	try:
		if ft == 1:
			# pibase_consensus
			ic = cols.index ('# C')
			cn = '# ContigName'
		elif ft == 2:
			# pibase_fisherdiff_bestchoice
			ic = cols.index ('Chr')
			cn = 'ContigName'
		else:
			# pibase_fisherdiff or pibase_diff
			ic = cols.index ('C')
			cn = 'ContigName'
		cols[ic] = cn
		headernew1 = ''
		for co in cols:
			headernew1 += co + '\t'
		headernew = headernew1.strip() + '\n'
		
	except:
			print ' *** Error: contig number not found in header line'
			print ' *** header line = ' + header
			sys.exit (' *** Aborted' )
	# print 'ic',ic


	############################################################################
	# open infile and outfile and write brief header lines to outfile:
	############################################################################
	
	inf = open (infile, 'r')
	out = open (outfile, 'w')
	
	outtime = time.strftime("%Y_%m_%d__%H:%M:%S") + "\n"
	out.write ('# File created by pibase_c_to_contig v0.9 on '+ outtime)
	out.write ('# from: %s\n#\n' %(infile))
	
	while True:
		line = inf.readline()
		if line.find('RefSeqContext\tGC[%]')> -1 :
			out.write(headernew)
			break
		else:
			out.write(line)
	# print line

	############################################################################
	# loop over remaining lines of infile
	############################################################################

	while True:
		line = inf.readline().strip()
		if len(line) == 0:
			# EOF
			break
		cols = line.split('\t')
		i=int(cols[ic])
		cols[ic] = chrom[i]
		lino=''
		for co in cols:
			lino += co + '\t'
		out.write(lino.strip() + '\n')

	out.close()
	inf.close()

##############################################
# the actual main
##############################################

if len(sys.argv) < 3:
	print pib_c_to_contig.__doc__
	sys.exit()
else:
	infile = sys.argv[1]
	if len(sys.argv) > 3:
		infile2 = sys.argv[2]
		outfile = sys.argv[3]
	else:
		infile2 = sys.argv[1]
		outfile = sys.argv[2]
		
setlimits()

pib_c_to_contig (infile, infile2, outfile)
print('Done')
