#!/usr/bin/env python
##!/usr/bin/python
##!/bin/env python2

import sys
import os.path
import time
import resource

###########################################################################################################
#
#  pibase_consensus - Add position information from multi-files into 1 file and infer genotypes for 
#                     10 different filter levels
#
###########################################################################################################

def setlimits():
	#############################################################################
	# sets resource limits (mandatory for job interoperability on cluster)
	# to 10MB Stack + 1GB data + 1GB RLIMIT_RSS
	# first steps taken from
	# http://docs.python.org/library/resource.html and
	# http://www.doughellmann.com/PyMOTW/resource/
	# Author: MICHAEL FORSTER m.forster@ikmb.uni-kiel.de
	# Version: 13 September 2010
	#############################################################################

	# usage = resource.getrusage(resource.RUSAGE_SELF)

	#print ' usage ='
	#print usage
	
	#for name, desc in [
	#('RLIMIT_CORE', 'core file size'),
	#('RLIMIT_CPU',  'CPU time'),
	#('RLIMIT_FSIZE', 'file size'),
	#('RLIMIT_DATA', 'heap size'),
	#('RLIMIT_STACK', 'stack size'),
	#('RLIMIT_RSS', 'resident set size'),
	#('RLIMIT_NPROC', 'number of processes'),
	#('RLIMIT_NOFILE', 'number of open files'),
	#('RLIMIT_MEMLOCK', 'lockable memory address'),
	#]:
	#	limit_num = getattr(resource, name)
	#	soft, hard = resource.getrlimit(limit_num)
	#	print 'Maximum %-25s (%-15s) : %20s %20s' % (desc, name, soft, hard)

	#############################################################################
	# Adjust the maximum size (in bytes) of the process's heap
	#############################################################################

	soft, hard = resource.getrlimit(resource.RLIMIT_DATA)
	#print 'Data/heap limit starts as  :', soft

	# 1GB heap:
	newheap=1024*1024*1024

	resource.setrlimit(resource.RLIMIT_DATA, (newheap, hard))

	soft, hard = resource.getrlimit(resource.RLIMIT_DATA)
	#print 'Data/heap limit changed to :', soft

	#############################################################################
	# ca 10MB original stack changed to exactly 10MB stack:
	#############################################################################

	#newstack=10*1024*1024+1
	newstack=10*1024*1024
	soft, hard = resource.getrlimit(resource.RLIMIT_STACK)
	#print 'Stack limit starts as  :', soft
	resource.setrlimit(resource.RLIMIT_STACK, (newstack, hard))
	soft, hard = resource.getrlimit(resource.RLIMIT_STACK)
	#print 'Stack limit changed to :', soft

	#############################################################################
	# 1GB RLIMIT_RSS
	# "Maximale resident set size in Byte. Bei Speicherengpaessen entzieht der Kern
	# den Prozessen, die den resident set size ueberschreiten, den zuviel
	# angeforderten Speicher."
	# http://pronix.linuxdelta.de/C/Linuxprogrammierung/Linuxsystemprogrammieren_C_Kurs_Kapitel3a.shtml
	#############################################################################

	newrss=1024*1024*1024
	soft, hard = resource.getrlimit(resource.RLIMIT_RSS)
	#print 'RLIMIT_RSS starts as  :', soft
	resource.setrlimit(resource.RLIMIT_RSS, (newrss, hard))
	soft, hard = resource.getrlimit(resource.RLIMIT_RSS)
	#print 'RLIMIT_RSS changed to :', soft

	#sys.exit()

def pib_identical_genotype (cgsmq, cgcmq, cgsmm, cgcmm, cgsm, cgcm, cgsq, cgcq, cgs, cgc, ncovs, nusps, 
	ncovs4, nusps4, refclass, th2, th4) :
	#############################################################################
	# Look at genotypes from all 10 filters; if they are identical, return this genotype, otherwise '?'
	# Must look at coverages and context too.
	#############################################################################
	cgeno = '?'
	# print ' * in pib_identical_genotype:  ncovs4, nusps4, refclass', ncovs4, nusps4, refclass
	if (cgsmq == cgcmq and cgsmq == cgsmm and cgsmq == cgcmm and cgsmq == cgsm and 
		cgsmq == cgcm and cgsmq == cgsq and cgsmq == cgcq and cgsmq == cgs and cgsmq == cgc):
		if nusps4 >= th4 and ncovs4 >= th2:
			cgeno = cgsmq
		elif nusps >= th4 and ncovs >= th2 and refclass < 5:
			cgeno = cgsmq
		elif cgsmq == '':
			cgeno = ''
	return cgeno
	
def pib_best_genotype (cgsmq, cgcmq, cgsmm, cgcmm, cgsm, cgcm, cgsq, cgcq, cgs, cgc,
			issmqa1, issmqa2, issmqc1, issmqc2, issmqg1, issmqg2, issmqt1, issmqt2,
			icmqa1, icmqa2, icmqc1, icmqc2, icmqg1, icmqg2, icmqt1, icmqt2,
			issmma1, issmma2, issmmc1, issmmc2, issmmg1, issmmg2, issmmt1, issmmt2,
			icmma1, icmma2, icmmc1, icmmc2, icmmg1, icmmg2, icmmt1, icmmt2,
			issma1, issma2, issmc1, issmc2, issmg1, issmg2, issmt1, issmt2,
			icma1, icma2, icmc1, icmc2, icmg1, icmg2, icmt1, icmt2,
			issqa1, issqa2, issqc1, issqc2, issqg1, issqg2, issqt1, issqt2,
			icqa1, icqa2, icqc1, icqc2, icqg1, icqg2, icqt1, icqt2,
			issa1, issa2, issc1, issc2, issg1, issg2, isst1, isst2, 
			ica1, ica2, icc1, icc2, icg1, icg2, ict1, ict2, refclass, th2, th4,
			homolog, hypervar, oldver, version) :
	#############################################################################
	# Decide on best genotype...
	#############################################################################
	
	# put genotypes into list:
	cg=[]
	cg.append(cgsmq)
	cg.append(cgcmq)
	cg.append(cgsmm)
	cg.append(cgcmm)
	cg.append(cgsm)
	cg.append(cgcm)
	cg.append(cgsq)
	cg.append(cgcq)
	cg.append(cgs)
	cg.append(cgc)
	#print 'cg: ', cg
	
	# get number of reads for each filter:
	n=[]
	n.append(pib_nreads(issmqa1, issmqa2, issmqc1, issmqc2, issmqg1, issmqg2, issmqt1, issmqt2) )
	n.append(pib_nreads(icmqa1, icmqa2, icmqc1, icmqc2, icmqg1, icmqg2, icmqt1, icmqt2) )
	n.append(pib_nreads(issmma1, issmma2, issmmc1, issmmc2, issmmg1, issmmg2, issmmt1, issmmt2) )
	n.append(pib_nreads(icmma1, icmma2, icmmc1, icmmc2, icmmg1, icmmg2, icmmt1, icmmt2) )
	n.append(pib_nreads(issma1, issma2, issmc1, issmc2, issmg1, issmg2, issmt1, issmt2) )
	n.append(pib_nreads(icma1, icma2, icmc1, icmc2, icmg1, icmg2, icmt1, icmt2) )
	n.append(pib_nreads(issqa1, issqa2, issqc1, issqc2, issqg1, issqg2, issqt1, issqt2) )
	n.append(pib_nreads(icqa1, icqa2, icqc1, icqc2, icqg1, icqg2, icqt1, icqt2) )
	n.append(pib_nreads(issa1, issa2, issc1, issc2, issg1, issg2, isst1, isst2) )
	n.append(pib_nreads(ica1, ica2, icc1, icc2, icg1, icg2, ict1, ict2) )
	#print 'n: ', n
	
	# define the minimum number of reads for each filter:
	m=[]
	m.append(int(th4))
	m.append(int(th2))
	m.append(int(th4))
	m.append(int(th2))
	m.append(int(th4))
	m.append(int(th2))
	m.append(int(th4))
	m.append(int(th2))
	m.append(int(th4))
	m.append(int(th2))
	
	nf=len(m)
	
	#####################################################
	# get majority consensus of the 10 genotypes, taking min reads into account:
	#####################################################
	
	# get genotype consensus numbers:
	cgm=[]
	ngm={}
	ngm_max=-1
	for i in range(0,nf):
		# print 'i, n[i], m[i]: ', i, n[i], m[i]
		if n[i] >= m[i]:
			if cg[i] in cgm:
				ngm[ cg[i] ] += 1
			else:
				cgm.append(cg[i])
				ngm[ cg[i] ] = 1
			ngm_max = max(ngm_max,ngm[ cg[i] ])
			
	# if consensus fails due to min reads threshold, repeat with nonzero reads:
	if ngm_max == -1 :
		for i in range(0,nf):
			# print 'i, n[i], m[i]: ', i, n[i], m[i]
			if n[i] >= 1:
				if cg[i] in cgm:
					ngm[ cg[i] ] += 1
				else:
					cgm.append(cg[i])
					ngm[ cg[i] ] = 1
				ngm_max = max(ngm_max,ngm[ cg[i] ])
	
	#print 'cgm: ', cgm
	#print 'ngm: ',ngm
	
	# majority consensus: 
	k=-1
	for i in range(0,nf):
		if ngm [ cgm[i] ] == ngm_max:
			k = i
			break
	cgmfinal = cgm[k]
	# print 'cgmfinal: ',cgmfinal
	
	#####################################################
	# get best-quality genotype that has "sufficient" reads, based on unique start-points:
	#####################################################
	
	cgb=''
	n1=-1
	for i in range(0,nf,2):
		if n[i] >= m[i]:
			cgb = cg[i]
			n1=i
			# print 'i, cgb, n1: ',i, cgb, n1
			break
	
	#####################################################
	# get best-quality supporting genotype that has "sufficient" reads, based on coverage:
	#####################################################
	
	cgbs=''
	n2=-1
	for i in range(1,nf,2):
		if n[i] >= m[i]:
			cgbs = cg[i]
			n2=i
			# print 'i, cgbs, n2: ',i, cgbs, n2
			break
	
	#####################################################
	# decide on genotype:
	#####################################################
	
	if oldver==1:
		# reverse compatibility mode: infer BestGen and BestQual using old rules
		if version == '1.4.3':
			if cgb == cgbs and n1>= 0:
				if refclass>4 and n1>3:
					# reference sequence context class looks difficult (homopolymeric run>4, or STRs) 
					# and mapping stringency was low
					if cgb == cgmfinal and ngm_max > 4:
						# this means that at least 4 filter-stages support this genotype.
						# >4 means than the least stringent ones (no filtering + basecall-qv filtering)
						# plus at least one "stringent" filter (e.g. mapped length) support this genotype
						return cgb,'?3'
					else:
						# this means that this genotype is supported by less than 5 filter stages, but
						# by at least 2 filter stages, of which one stage is in the unique start points category, 
						# and the other stage is in the coverage category
						return cgb,'?4'
				else:
					# reference sequence context class looks ok or mapping stringency is ok (e.g. 1MM in 49mer or 50mer)
					if cgb == cgmfinal and ngm_max > 4:
						# this means that at least 4 filter-stages support this genotype.
						# >4 means than the least stringent ones (no filtering + basecall-qv filtering)
						# plus at least one "stringent" filter (e.g. mapped length) support this genotype
						return cgb,'?1'
					else:
						# this means that this genotype is supported by less than 5 filter stages, but
						# by at least 2 filter stages, of which one stage is in the unique start points category, 
						# and the other stage is in the coverage category
						return cgb,'?2'
					
			else:
				# this means that either there is no start-point-derived genotype, or that there is a conflict between 
				# the best unique-start-point derived genotype and the best coverage-derived genotype
				if n1>= 0 and n1< n2:
					# the unique-start-point derived genotype is of senior quality to the genotype-derived genotype
					return cgb,'?5'
				elif n2 >= 0:
					# the coverage-derived genotype is of senior quality to the u.s.p.-derived genotype
					return cgbs,'?6'
				else:
					# coverage is below min threshold: this is a low-coverage genotype GUESS....
					if refclass<=4:
						return cgmfinal, '?7'
					else:
						return cgmfinal, '?8'
	else:
		# infer BestGen and Bestqual using the most current rules
		
			#####################################################
			# start off the same as v1.4.3:
			#####################################################
		
			if cgb == cgbs and n1>= 0:
				if refclass>4 and n1>3:
					# reference sequence context class looks difficult (homopolymeric run>4, or STRs) 
					# and mapping stringency was low
					if cgb == cgmfinal and ngm_max > 4:
						# this means that at least 4 filter-stages support this genotype.
						# >4 means than the least stringent ones (no filtering + basecall-qv filtering)
						# plus at least one "stringent" filter (e.g. mapped length) support this genotype
						bestgen = cgb
						bestqual ='?3'
					else:
						# this means that this genotype is supported by less than 5 filter stages, but
						# by at least 2 filter stages, of which one stage is in the unique start points category, 
						# and the other stage is in the coverage category
						bestgen = cgb
						bestqual ='?4'
				else:
					# reference sequence context class looks ok or mapping stringency is ok (e.g. 1MM in 49mer or 50mer)
					if cgb == cgmfinal and ngm_max > 4:
						# this means that at least 4 filter-stages support this genotype.
						# >4 means than the least stringent ones (no filtering + basecall-qv filtering)
						# plus at least one "stringent" filter (e.g. mapped length) support this genotype
						bestgen = cgb
						bestqual ='?1'
					else:
						# this means that this genotype is supported by less than 5 filter stages, but
						# by at least 2 filter stages, of which one stage is in the unique start points category, 
						# and the other stage is in the coverage category
						bestgen = cgb
						bestqual ='?2'
					
			else:
				# this means that either there is no start-point-derived genotype, or that there is a conflict between 
				# the best unique-start-point derived genotype and the best coverage-derived genotype
				if n1>= 0 and n1< n2:
					# the unique-start-point derived genotype is of senior quality to the genotype-derived genotype
					bestgen = cgb
					bestqual ='?5'
				elif n2 >= 0:
					# the coverage-derived genotype is of senior quality to the u.s.p.-derived genotype
					bestgen = cgbs
					bestqual ='?6'
				else:
					# coverage is below min threshold: this is a low-coverage genotype GUESS....
					bestgen = cgmfinal
					if refclass<=4:
						bestqual ='?7'
					else:
						bestqual ='?8'
			
			#####################################################
			# v1.4.4: Postprocessing for hypervariable regions (more than 1 SNP per read)
			#####################################################
			
			if bestqual != '?1':
				# exclude all '?1'-genotypes (quite high-quality) from the hypervariable site postprocessing
				if hypervar == 'V' :
					# In this hypervariable site, replace the previously computed BestGen genotype 
					# with the Filt1 genotype (base-QV-only filter, default=20).
					# The filter-level Filt1 genotype gives slightly better results than 
					# the filter-level Filt2 genotype (read-length-limit-filter),
					# because some full-length reads get truncated at indels, 
					# leading to lower coverage and sometimes to the required allele
					# of a heterozygous SNP being eliminated.
					
					if cgsq != '' and len(cgsq) == 2:
						bestgen = cgsq
			
			return bestgen, bestqual
	# end of routine

def pib_homolog_hypervar (issmqa1, 
	issmqa2, issmqc1, issmqc2, issmqg1, issmqg2, issmqt1, issmqt2,
	icmqa1, icmqa2, icmqc1, icmqc2, icmqg1, icmqg2, icmqt1, icmqt2,
	issmma1, issmma2, issmmc1, issmmc2, issmmg1, issmmg2, issmmt1, issmmt2,
	icmma1, icmma2, icmmc1, icmmc2, icmmg1, icmmg2, icmmt1, icmmt2,
	issma1, issma2, issmc1, issmc2, issmg1, issmg2, issmt1, issmt2,
	icma1, icma2, icmc1, icmc2, icmg1, icmg2, icmt1, icmt2,
	issqa1, issqa2, issqc1, issqc2, issqg1, issqg2, issqt1, issqt2,
	icqa1, icqa2, icqc1, icqc2, icqg1, icqg2, icqt1, icqt2,
	issa1, issa2, issc1, issc2, issg1, issg2, isst1, isst2, 
	ica1, ica2, icc1, icc2, icg1, icg2, ict1, ict2, th6, th7) :
	
	#####################################################
	# v1.4.4: Compute whether coord lies in a homologous
	# region (reads do not map uniquely to the reference).
	# and/or in a hypervariable region (more than 1 SNP
	# per read)
	#####################################################
	
	# get number of reads for each filter:
	n=[]
	n.append(pib_nreads(issmqa1, issmqa2, issmqc1, issmqc2, issmqg1, issmqg2, issmqt1, issmqt2) )
	n.append(pib_nreads(icmqa1, icmqa2, icmqc1, icmqc2, icmqg1, icmqg2, icmqt1, icmqt2) )
	n.append(pib_nreads(issmma1, issmma2, issmmc1, issmmc2, issmmg1, issmmg2, issmmt1, issmmt2) )
	n.append(pib_nreads(icmma1, icmma2, icmmc1, icmmc2, icmmg1, icmmg2, icmmt1, icmmt2) )
	n.append(pib_nreads(issma1, issma2, issmc1, issmc2, issmg1, issmg2, issmt1, issmt2) )
	n.append(pib_nreads(icma1, icma2, icmc1, icmc2, icmg1, icmg2, icmt1, icmt2) )
	n.append(pib_nreads(issqa1, issqa2, issqc1, issqc2, issqg1, issqg2, issqt1, issqt2) )
	n.append(pib_nreads(icqa1, icqa2, icqc1, icqc2, icqg1, icqg2, icqt1, icqt2) )
	n.append(pib_nreads(issa1, issa2, issc1, issc2, issg1, issg2, isst1, isst2) )
	n.append(pib_nreads(ica1, ica2, icc1, icc2, icg1, icg2, ict1, ict2) )
			
	if n[2]-n[0] >= max ( (n[2]+n[0])/2 * th6, th7) :
		# Filtering indicates that this SNP lies in a homologous site.
		# Reason: The difference in unique-start-point counts is significant
		# between filter level Filt4 (only uniquely mapping reads) 
		# and filter level Filt3 (may include non-uniquely mapping reads). 
		# The significance is computed using thresholds th6 and th7.
		homolog = 'H'
	else:
		homolog = ''
	
	if n[4]-n[2] >= max ( (n[4]+n[2])/2 * th6, th7) :
		# Filtering indicates that this SNP lies in a hypervariable site.
		# Reason: The difference in unique-start-point counts is significant
		# between filter level Filt3 (mismatches-per-read-limit, default=1) 
		# and filter level sFilt2 (no mismatches-per-read-limit). 
		# The significance is computed using thresholds th6 and th7.
		hypervar = 'V'
	else:
		hypervar = ''
	return homolog, hypervar


def pib_forward_reverse_confirm (cgbest,
	issmqa1, issmqa2, issmqc1, issmqc2, issmqg1, issmqg2, issmqt1, issmqt2,
	icmqa1, icmqa2, icmqc1, icmqc2, icmqg1, icmqg2, icmqt1, icmqt2,
	issmma1, issmma2, issmmc1, issmmc2, issmmg1, issmmg2, issmmt1, issmmt2,
	icmma1, icmma2, icmmc1, icmmc2, icmmg1, icmmg2, icmmt1, icmmt2,
	issma1, issma2, issmc1, issmc2, issmg1, issmg2, issmt1, issmt2,
	icma1, icma2, icmc1, icmc2, icmg1, icmg2, icmt1, icmt2,
	issqa1, issqa2, issqc1, issqc2, issqg1, issqg2, issqt1, issqt2,
	icqa1, icqa2, icqc1, icqc2, icqg1, icqg2, icqt1, icqt2,
	issa1, issa2, issc1, issc2, issg1, issg2, isst1, isst2, 
	ica1, ica2, icc1, icc2, icg1, icg2, ict1, ict2, cla, th5 ) :
	#######################################################################################
	# Check whether alleles A and B of best genotype are confirmed in forward and reverse read directions
	# Threshold th5: read direction with least read counts must have at least e.g. 20% of 
	# counts of reads of other read direction;
	# e.g.: forward: 5 counts, reverse 1 count --> confirmed allele 
	#######################################################################################
	
	# for easier looping, define a nested list of tuples:
	
	a = [ (issmqa1, issmqa2), (icmqa1, icmqa2), (issmma1, issmma2), (icmma1, icmma2), 
	(issma1, issma2), (icma1, icma2), (issqa1, issqa2), (icqa1, icqa2), (issa1, issa2), (ica1, ica2) ]
	
	c = [ (issmqc1, issmqc2), (icmqc1, icmqc2), (issmmc1, issmmc2), (icmmc1, icmmc2), 
	(issmc1, issmc2), (icmc1, icmc2), (issqc1, issqc2), (icqc1, icqc2), (issc1, issc2), (icc1, icc2) ]
	
	g = [ (issmqg1, issmqg2), (icmqg1, icmqg2), (issmmg1, issmmg2), (icmmg1, icmmg2), 
	(issmg1, issmg2), (icmg1, icmg2), (issqg1, issqg2), (icqg1, icqg2), (issg1, issg2), (icg1, icg2) ]
	
	t = [ (issmqt1, issmqt2), (icmqt1, icmqt2), (issmmt1, issmmt2), (icmmt1, icmmt2), 
	(issmt1, issmt2), (icmt1, icmt2), (issqt1, issqt2), (icqt1, icqt2), (isst1, isst2), (ict1, ict2) ]
	
	acgt = [ a, c, g, t]
	
	# define a dictionary to access the nested list:
	d = {}
	d['A'] = 0
	d['C'] = 1
	d['G'] = 2
	d['T'] = 3
	
	# list of indexes:
	k = []
	try:
		k.append ( d [cgbest[:1]] )
		k.append ( d [cgbest[1:]] )
	except:
		print '*** WARNING: More than 2 "best" alleles in genotype:'
		print '*** cgbest = ',cgbest
		#print '*** cgbest[:1] = ',cgbest[:1]
		#print '*** cgbest[1:] = ',cgbest[1:]
	#print 'k=',k
	
	# forward/reverse confirmation marker:
	m=[]
	
	# loop over nested list with appropriate index :
	for i in k:
		# print acgt [i]
		m1=''
		m2=''
		for ii in acgt [i]:
			# print ii
			f,r = ii
			#print 'f, r = ', f,',',r
			#tmin = max( 1.0 , 0.2 * max(f,r) )
			tmin = max( 1.0 , th5 * max(f,r) )
			#print 'tmin=',tmin
			if f >= tmin:
				m1='+'
			if r >= tmin:
				m2='-'
		m.append(m1+m2)
	return m[0], m[1]

def pib_minreads (a1,a2, c1,c2, g1,g2, t1,t2, th1):
	#############################################################################
	# Consensus genotype from counts of A+/A-, C+/C-, G+/G-, T+/T-
	# where + and - denote the pos and neg strand
	# and counts denote the number of reads.
	#############################################################################
	total = a1+a2+ c1+c2+ g1+g2+ t1+t2
	#min = max(int(total*0.022),1)
	min = max(int(total*th1),1)
	return min

def pib_minusps (a1,a2, c1,c2, g1,g2, t1,t2, th3):
	#############################################################################
	# Consensus genotype from counts of A+/A-, C+/C-, G+/G-, T+/T-
	# where + and - denote the pos and neg strand
	# and counts denote the number of unique start points.
	#############################################################################
	total = a1+a2+ c1+c2+ g1+g2+ t1+t2
	min = max(int(total*th3),1)
	return min

def pib_nreads (a1,a2, c1,c2, g1,g2, t1,t2):
	#############################################################################
	# Total counts of A+/A-, C+/C-, G+/G-, T+/T-
	# where + and - denote the pos and neg strand
	# and counts denote the number of reads.
	#############################################################################
	total = a1+a2+ c1+c2+ g1+g2+ t1+t2
	return total

def pib_genotype (a1,a2, c1,c2, g1,g2, t1,t2, min):
	#############################################################################
	# Consensus genotype from counts of A+/A-, C+/C-, G+/G-, T+/T-
	# where + and - denote the pos and neg strand
	# and counts denote the number of reads.
	#############################################################################
	
	total = pib_nreads (a1,a2, c1,c2, g1,g2, t1,t2)
	if total <= min:
		min1 = 0
	else:
		min1 = min
	#print 'total, min1, min: ', total, min1, min
		
	c=''
	if (a1+a2 > min1):
		c+='A'
	if (c1+c2 > min1):
		c+='C'
	if (g1+g2 > min1):
		c+='G'
	if (t1+t2 > min1):
		c+='T'
	if len(c)== 1:
		c+=c
	n='%s:%s:%s:%s' % (a1+a2, c1+c2, g1+g2, t1+t2)
	#print 'c,n: ', c,n
	return c, n

def pib_consensus (args, nargs, outfile, th1, th2, th3, th4, th5, th6, th7, oldver, version):
	'''        #######################################################################################
	#  pibase_consensus v1.4.5
	#  add up values in pibase-output-files into one file, estimate "best" genotype,
	#  flag the "BestGen" with "?" if it is instable with respect to filtering, 
	#  flag hypervariable loci. 
	#  Flag homologous/ambiguous loci. Note that this flag is informative only 
	#  if all non-uniquely-mapped reads were available in the BAM file.   
	#######################################################################################
	#  pibase_consensus [-t[h][v [ver]] t1 t2 t3 t4 t5 {h1 h2}] in1 [in2 ... inN] out
	#
	#   []=optional   {}=only for "-th" option
	#
	#   [-t[h][v]  (optional) minimal thresholds for allele calling (A, C, G, or T),
	#              (optionally) followed by "h" to override h1 and h2 defaults,
	#              (optionally) followed by "v" to revert to an older version
	#
	#     [ver]    (optional) older pibase_consensus version to revert to.
	#              To list all available older versions, specify "-tv" without ver
	#              (older version 1.4.3: arguments h1 and h2 not available) 
	#              
	#      t1      min fraction of reads indicating allele (default: 2.2%, i.e. 0.022)
	#      t2      min number of reads indicating allele   (default: 8)
	#      t3      min fraction of unique start points indicating allele (default: 0.04)
	#      t4      min number of unique start points indicating allele   (default: 4)
	#      t5      both-stranded confirmation only if, for at least one filter level,
	#              min-strand-counts >= t5 * max-strand-counts (default t5: 0.2)
	#
	#     {h1      min fraction of unique start points that need to be different between
	#              filter F2-F3 resp. F3-F4 in order to flag a hypervariable resp. 
	#              homologous locus (default: 2.2%, i.e. 0.022)
	#
	#      h2} ]   min number of unique start points difference between F2-F3 resp. F3-F4
	#              in order to flag a hypervariable resp. homologous locus (default: 2)
	#
	#    inX       input file(s) = pibase_bamref-output file(s);
	#              must use same reference sequence (NOT CHECKED!)
	#              and same positions of interest
	#
	#    out       name for (tab-separated) output file
	#######################################################################################
	#  Author: Michael Forster                                 m.forster@ikmb.uni-kiel.de
	#  2011/10/27, 2011/12/09, 2012/05/21
	#######################################################################################
	'''

	#######################################################################################
	# open infiles and output file
	#######################################################################################

	inf=[]
	#print 'nargs=', nargs
	#print args
	for i in range(1,nargs-1):
		# print 'infile i =',i
		inff=open(args[i], "r")
		# print 'infile: i =',i,' name=',args[i]
		print 'infile:  ',args[i]
		inf.append(inff)
	
	out = open(outfile,"w")
	print 'outfile: ', outfile


	#######################################################################################################
	#  Write HEADER of OUT FILE and read all header lines of 1st input file
	#######################################################################################################

	# out.write ('# Position Info (1-based coordinates) generated by pibase from file '+ sam_filename+'\n')
	out.write ('# Estimate of Best Genotype and its quality (BestQual tag), for consensus of pibase-generated Position Info Files (1-based coordinates)\n')
	# out.write ('# Generated by pibase_consensus.py v1.4.4 from files\n')
	out.write ('# Generated by pibase_consensus v%s from files\n' % (version))
	for i in range(1,nargs-1):
		out.write ('# '+args[i]+'\n')
		
	out.write ('# Using coverage thresholds t1=%s t2=%s, unique start point thresholds t3=%s t4=%s, and both-strand-confirmation threshold t5=%s\n' % (th1, th2, th3, th4, th5))
	if version != '1.4.3':
		out.write ('# Using unique start point differences t6=%s t7=%s between filter levels F3-F2 resp. F4-F3 for flagging hypervariable loci resp. homologous loci\n' % (th6, th7))
	out.write('#\n')

	# Read header line from first pibase-output file
	line = inf[0].readline()

	while True:
		line=inf[0].readline()
		# print 'line ==== '+line
		if len(line) == 0:
			# EOF reached
			sys.exit('Error: unexpected end of file in header lines')
		else:
			line = line.strip()
			
			if line.startswith("# Date:"):
				continue
			elif line.startswith("# C\tPos"):
				# print 'break at line: '+line
				break
			elif line.startswith("# \t") and version == '1.4.3':
				# print 'continue at line: '+line
				continue
			elif line.startswith("# "):
				# exclude two pibase_bamref header-lines which 
				# are correctly re-generated by pibase_consensus:
				if line.find("MapQV>RQVmin and MM<=MMmax") > -1:
					continue
				if line.find("Unique start points (on +- strands)") > -1:
					continue
				
				out.write(line+'\n')
			elif  line == '':
				out.write('#\n')
			else:
				# just as a catch-all:
				break
				
	outtime = time.strftime("%Y_%m_%d__%H:%M:%S") + "\n#"
	out.write ('# Date: '+ outtime+'\n')
	out.write ('# Column "BestGen" is the best estimate of the genotype using all filter levels, coverages, and start points\n')
	out.write ('# Column "BestQual" contains a "?" if the genotype estimate is of low quality, followed by a number from 1 (slight quality loss) to 8 (low coverage guess)\n')
	if version == '1.4.3':
		out.write ('# Columns "A+-" (allele 1) and "B+-" (allele 2) contain "+-" if both strands support the allele (strand defined as supported if reads on "minor" strand <= 0.2 * reads on "major" strand), otherwise just "+" or "-"\n#\n')
	else:
		out.write ('# Columns "A+-" (allele 1) and "B+-" (allele 2) contain "+-" if both strands support the allele (strand defined as supported if reads on "minor" strand <= 0.2 * reads on "major" strand), otherwise just "+" or "-"\n')
		out.write ('# Column "Homologous" contains "H" if the coordinate lies in a region of the reference sequence where some reads have not mapped uniquely\n')
		out.write ('# Column "Hypervariable" contains "V" if the coordinate lies in a hypervariable region, where there is more than 1 mismatch per read.\n#\n')

	### Later at this stage: consensus genotypes in first columns

	if version == '1.4.3':
		### Old-version-compatibility: No columns "Homologous" or "Hypervariable"
		c = '# \t\t\t\t\t\t' + \
		'\tConsensus Genotype\t\t\t\t\t\t\t\t\t' + \
		'\t\t' + \
		'\t\t' + \
		'\tReads\t\t\t\t\t\t\t\t\t' + \
		'\tMapQV>RQVmin and MM<=MMmax and Read>MLmin and BaseQVs>=QVmin\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t' + \
		'\tMM<=MMmax and Read>MLmin and BaseQVs>=QVmin\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t' + \
		'\tRead>MLmin and BaseQVs>=QVmin\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t' + \
		'\tBaseQVs>=QVmin\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t' + \
		'\tNo QV-Filtering\n'
		out.write (c)
		# out.write ('# \t\t\t\t\t\t\tUnique start points (on +- strands)\t\t\t\t\t\t\t\t\t\tCoverage\t\t\t\t\t\t\t\t\t\tUnique start points (on +- strands)\t\t\t\t\t\t\t\t\t\tCoverage\t\t\t\t\t\t\t\t\t\tUnique start points (on +- strands)\t\t\t\t\t\t\t\t\t\tCoverage\t\t\t\t\t\t\t\t\t\tUnique start points\t\t\t\t\t\t\t\t\t\tCoverage\t\t\t\t\t\t\t\t\t\n')
		c = '# \t\t\t\t\t\t' + \
		'\tBest\t' +\
		'\t\t' + \
		'\tFilt4\t\tFilt3\t\tFilt2\t\tFilt1\t\tFilt0\t' +\
		'\tFilt4\t\tFilt3\t\tFilt2\t\tFilt1\t\tFilt0\t' +\
		'\tUnique start points (on +- strands)\t\t\t\t\t\t\t\t\t\tCoverage\t\t\t\t\t\t\t\t\t' + \
		'\tUnique start points (on +- strands)\t\t\t\t\t\t\t\t\t\tCoverage\t\t\t\t\t\t\t\t\t' + \
		'\tUnique start points (on +- strands)\t\t\t\t\t\t\t\t\t\tCoverage\t\t\t\t\t\t\t\t\t' + \
		'\tUnique start points (on +- strands)\t\t\t\t\t\t\t\t\t\tCoverage\t\t\t\t\t\t\t\t\t' + \
		'\tUnique start points (on +- strands)\t\t\t\t\t\t\t\t\t\tCoverage\t\t\t\t\t\t\t\t\t\n'
		out.write (c)
		# out.write ('# C\tPos[1]\tRef\tRefSeqContext\tGC[%]\tClass\tIgn\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\n')
		c = '# C\tPos[1]\tRef\tRefSeqContext\tGC[%]\tClass\tIgn' + \
			'\tBestGen\tBestQual\tA+-\tB+-' + \
			'\tSP\tCV\tSP\tCV\tSP\tCV\tSP\tCV\tSP\tCV'+ \
			'\tSP\tCV\tSP\tCV\tSP\tCV\tSP\tCV\tSP\tCV'+ \
			'\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-'+ \
			'\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-'+ \
			'\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-'+ \
			'\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-'+ \
			'\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\n'
		out.write (c)
	else:
		### Current version: Additional columns "Homologous" or "Hypervariable"
		c = '# \t\t\t\t\t\t' + \
		'\tConsensus Genotype\t\t\t\t\t\t\t\t\t' + \
		'\t\t' + \
		'\t\t' + \
		'\tReads\t\t\t\t\t\t\t\t\t' + \
		'\t\t'+\
		'\tMapQV>RQVmin and MM<=MMmax and Read>MLmin and BaseQVs>=QVmin\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t' + \
		'\tMM<=MMmax and Read>MLmin and BaseQVs>=QVmin\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t' + \
		'\tRead>MLmin and BaseQVs>=QVmin\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t' + \
		'\tBaseQVs>=QVmin\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t' + \
		'\tNo QV-Filtering\n'
		out.write (c)
		# out.write ('# \t\t\t\t\t\t\tUnique start points (on +- strands)\t\t\t\t\t\t\t\t\t\tCoverage\t\t\t\t\t\t\t\t\t\tUnique start points (on +- strands)\t\t\t\t\t\t\t\t\t\tCoverage\t\t\t\t\t\t\t\t\t\tUnique start points (on +- strands)\t\t\t\t\t\t\t\t\t\tCoverage\t\t\t\t\t\t\t\t\t\tUnique start points\t\t\t\t\t\t\t\t\t\tCoverage\t\t\t\t\t\t\t\t\t\n')
		c = '# \t\t\t\t\t\t' + \
		'\tBest\t' +\
		'\t\t' + \
		'\tFilt4\t\tFilt3\t\tFilt2\t\tFilt1\t\tFilt0\t' +\
		'\tFilt4\t\tFilt3\t\tFilt2\t\tFilt1\t\tFilt0\t' +\
		'\t\t'+\
		'\tUnique start points (on +- strands)\t\t\t\t\t\t\t\t\t\tCoverage\t\t\t\t\t\t\t\t\t' + \
		'\tUnique start points (on +- strands)\t\t\t\t\t\t\t\t\t\tCoverage\t\t\t\t\t\t\t\t\t' + \
		'\tUnique start points (on +- strands)\t\t\t\t\t\t\t\t\t\tCoverage\t\t\t\t\t\t\t\t\t' + \
		'\tUnique start points (on +- strands)\t\t\t\t\t\t\t\t\t\tCoverage\t\t\t\t\t\t\t\t\t' + \
		'\tUnique start points (on +- strands)\t\t\t\t\t\t\t\t\t\tCoverage\t\t\t\t\t\t\t\t\t\n'
		out.write (c)
		# out.write ('# C\tPos[1]\tRef\tRefSeqContext\tGC[%]\tClass\tIgn\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\n')
		c = '# C\tPos[1]\tRef\tRefSeqContext\tGC[%]\tClass\tIgn' + \
			'\tBestGen\tBestQual\tA+-\tB+-' + \
			'\tSP\tCV\tSP\tCV\tSP\tCV\tSP\tCV\tSP\tCV'+ \
			'\tSP\tCV\tSP\tCV\tSP\tCV\tSP\tCV\tSP\tCV'+ \
			'\tHomologous\tHypervariable'+\
			'\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-'+ \
			'\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-'+ \
			'\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-'+ \
			'\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-'+ \
			'\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\n'
		out.write (c)

	#######################################################################################################
	#  Read all header lines of 2nd to nth input files
	#######################################################################################################

	for i in range(2,nargs-1):
		j=i-1
		while True:
			line=inf[j].readline()
			if len(line) == 0:
				# EOF reached
				sys.exit('Error: unexpected end of file in header lines')
			else:
				line = line.strip()
				if line.startswith("# C\tPos"):
					break
				elif line.startswith("# ") or line == '':
					continue
				else:
					# just as a catch-all:
					break

	#######################################################################################################
	# Start of WHILE loop over all positions of interest
	#######################################################################################################

	while True:

		#######################################################################################
		# Read data from first file (esp. poi, to check that all other file's pois are ok)
		#######################################################################################

		line=inf[0].readline()
		if len(line) == 0:
			# EOF reached
			break

		# for debug: outfile
		# out.write (line)

		line = line.strip()

		co = line.split('\t')

		# Should be identical for all files:
		try:
			ichrom = int(co[0])
		except:
			print 'Error in first file: '+args[1]
			print 'No ichrom in line: '+line+'\n'
			sys.exit()
			
		poi = int(co[1])
		refbasepoi = co[2]
		ctxt = co[3]
		gcc = int(co[4])
		cla = int(co[5])

		# Items to be added with the corresponding items of the other files (quick and dirty for clarity):
		indelread = int(co[6])

		issmqa1 = int(co[7])
		issmqa2 = int(co[8])
		issmqc1 = int(co[9])
		issmqc2 = int(co[10])
		issmqg1 = int(co[11])
		issmqg2 = int(co[12])
		issmqt1 = int(co[13])
		issmqt2 = int(co[14])
		issmqn1 = int(co[15])
		issmqn2 = int(co[16])

		icmqa1 = int(co[17])
		icmqa2 = int(co[18])
		icmqc1 = int(co[19])
		icmqc2 = int(co[20])
		icmqg1 = int(co[21])
		icmqg2 = int(co[22])
		icmqt1 = int(co[23])
		icmqt2 = int(co[24])
		icmqn1 = int(co[25])
		icmqn2 = int(co[26])

		issmma1 = int(co[27])
		issmma2 = int(co[28])
		issmmc1 = int(co[29])
		issmmc2 = int(co[30])
		issmmg1 = int(co[31])
		issmmg2 = int(co[32])
		issmmt1 = int(co[33])
		issmmt2 = int(co[34])
		issmmn1 = int(co[35])
		issmmn2 = int(co[36])

		icmma1 = int(co[37])
		icmma2 = int(co[38])
		icmmc1 = int(co[39])
		icmmc2 = int(co[40])
		icmmg1 = int(co[41])
		icmmg2 = int(co[42])
		icmmt1 = int(co[43])
		icmmt2 = int(co[44])
		icmmn1 = int(co[45])
		icmmn2 = int(co[46])

		issma1 = int(co[47])
		issma2 = int(co[48])
		issmc1 = int(co[49])
		issmc2 = int(co[50])
		issmg1 = int(co[51])
		issmg2 = int(co[52])
		issmt1 = int(co[53])
		issmt2 = int(co[54])
		issmn1 = int(co[55])
		issmn2 = int(co[56])

		icma1 = int(co[57])
		icma2 = int(co[58])
		icmc1 = int(co[59])
		icmc2 = int(co[60])
		icmg1 = int(co[61])
		icmg2 = int(co[62])
		icmt1 = int(co[63])
		icmt2 = int(co[64])
		icmn1 = int(co[65])
		icmn2 = int(co[66])

		issqa1 = int(co[67])
		issqa2 = int(co[68])
		issqc1 = int(co[69])
		issqc2 = int(co[70])
		issqg1 = int(co[71])
		issqg2 = int(co[72])
		issqt1 = int(co[73])
		issqt2 = int(co[74])
		issqn1 = int(co[75])
		issqn2 = int(co[76])

		icqa1 = int(co[77])
		icqa2 = int(co[78])
		icqc1 = int(co[79])
		icqc2 = int(co[80])
		icqg1 = int(co[81])
		icqg2 = int(co[82])
		icqt1 = int(co[83])
		icqt2 = int(co[84])
		icqn1 = int(co[85])
		icqn2 = int(co[86])

		issa1 = int(co[87])
		issa2 = int(co[88])
		issc1 = int(co[89])
		issc2 = int(co[90])
		issg1 = int(co[91])
		issg2 = int(co[92])
		isst1 = int(co[93])
		isst2 = int(co[94])
		issn1 = int(co[95])
		issn2 = int(co[96])
		
		ica1 = int(co[97])
		ica2 = int(co[98])
		icc1 = int(co[99])
		icc2 = int(co[100])
		icg1 = int(co[101])
		icg2 = int(co[102])
		ict1 = int(co[103])
		ict2 = int(co[104])
		icn1 = int(co[105])
		icn2 = int(co[106])

		#######################################################################################
		# Read data from 2nd to nth file, checking that poi is identical to 1st file's poi
		#######################################################################################

		for i in range(2,nargs-1):
			j=i-1
			line=inf[j].readline()
			if len(line) == 0:
				# EOF reached
				sys.exit('Error: unexpected end of file: '+ args[i])

			# for debug: outfile
			# out.write (line)

			line = line.strip()

			co = line.split('\t')

			# Should be identical for all files:
			t_ichrom = int(co[0])
			t_poi = int(co[1])
			t_refbasepoi = co[2]
			t_ctxt = co[3]
			t_gcc = int(co[4])
			t_cla = int(co[5])

			if (t_ichrom!=ichrom or t_poi!=poi or t_refbasepoi!=refbasepoi or t_ctxt!=ctxt or t_gcc!=gcc or t_cla!=cla):
				c='Error: File %s does not correspond to file %s \n' % (args[1], args[2])
				c+='chrom %s / chrom %s \n' %(ichrom, t_ichrom)
				c+='pos %s / pos %s \n' %(poi, t_poi)
				c+='refbase %s / refbase %s \n' %(refbasepoi, t_refbasepoi)
				c+='context %s / context %s \n' %(ctxt, t_ctxt)
				c+='GC% %s / GC% %s \n' %(gcc, t_gcc)
				c+='ContextClass %s / ContextClass  %s \n' %(cla, t_cla)
				sys.exit(c)

			indelread += int(co[6])

			issmqa1 += int(co[7])
			issmqa2 += int(co[8])
			issmqc1 += int(co[9])
			issmqc2 += int(co[10])
			issmqg1 += int(co[11])
			issmqg2 += int(co[12])
			issmqt1 += int(co[13])
			issmqt2 += int(co[14])
			issmqn1 += int(co[15])
			issmqn2 += int(co[16])

			icmqa1 += int(co[17])
			icmqa2 += int(co[18])
			icmqc1 += int(co[19])
			icmqc2 += int(co[20])
			icmqg1 += int(co[21])
			icmqg2 += int(co[22])
			icmqt1 += int(co[23])
			icmqt2 += int(co[24])
			icmqn1 += int(co[25])
			icmqn2 += int(co[26])

			issmma1 += int(co[27])
			issmma2 += int(co[28])
			issmmc1 += int(co[29])
			issmmc2 += int(co[30])
			issmmg1 += int(co[31])
			issmmg2 += int(co[32])
			issmmt1 += int(co[33])
			issmmt2 += int(co[34])
			issmmn1 += int(co[35])
			issmmn2 += int(co[36])

			icmma1 += int(co[37])
			icmma2 += int(co[38])
			icmmc1 += int(co[39])
			icmmc2 += int(co[40])
			icmmg1 += int(co[41])
			icmmg2 += int(co[42])
			icmmt1 += int(co[43])
			icmmt2 += int(co[44])
			icmmn1 += int(co[45])
			icmmn2 += int(co[46])

			issma1 += int(co[47])
			issma2 += int(co[48])
			issmc1 += int(co[49])
			issmc2 += int(co[50])
			issmg1 += int(co[51])
			issmg2 += int(co[52])
			issmt1 += int(co[53])
			issmt2 += int(co[54])
			issmn1 += int(co[55])
			issmn2 += int(co[56])

			icma1 += int(co[57])
			icma2 += int(co[58])
			icmc1 += int(co[59])
			icmc2 += int(co[60])
			icmg1 += int(co[61])
			icmg2 += int(co[62])
			icmt1 += int(co[63])
			icmt2 += int(co[64])
			icmn1 += int(co[65])
			icmn2 += int(co[66])

			issqa1 += int(co[67])
			issqa2 += int(co[68])
			issqc1 += int(co[69])
			issqc2 += int(co[70])
			issqg1 += int(co[71])
			issqg2 += int(co[72])
			issqt1 += int(co[73])
			issqt2 += int(co[74])
			issqn1 += int(co[75])
			issqn2 += int(co[76])

			icqa1 += int(co[77])
			icqa2 += int(co[78])
			icqc1 += int(co[79])
			icqc2 += int(co[80])
			icqg1 += int(co[81])
			icqg2 += int(co[82])
			icqt1 += int(co[83])
			icqt2 += int(co[84])
			icqn1 += int(co[85])
			icqn2 += int(co[86])

			issa1 += int(co[87])
			issa2 += int(co[88])
			issc1 += int(co[89])
			issc2 += int(co[90])
			issg1 += int(co[91])
			issg2 += int(co[92])
			isst1 += int(co[93])
			isst2 += int(co[94])
			issn1 += int(co[95])
			issn2 += int(co[96])

			ica1 += int(co[97])
			ica2 += int(co[98])
			icc1 += int(co[99])
			icc2 += int(co[100])
			icg1 += int(co[101])
			icg2 += int(co[102])
			ict1 += int(co[103])
			ict2 += int(co[104])
			icn1 += int(co[105])
			icn2 += int(co[106])

		#######################################################################################
		# Threshold of reads, below which a sequencing error might have happened (e.g. 1 read, or 1.5% of coverage at this position)
		#######################################################################################
		
		# For coverages:
		min = pib_minreads (ica1, ica2, icc1, icc2, icg1, icg2, ict1, ict2, th1)
		# print 'min = ', min
		
		# For unique start-points:
		mins = pib_minusps (issa1, issa2, issc1, issc2, issg1, issg2, isst1, isst2, th3)
		# print 'mins = ', mins
		
		# For coverages:
		min1 = pib_minreads (icqa1, icqa2, icqc1, icqc2, icqg1, icqg2, icqt1, icqt2, th1)
		# print 'min = ', min
		
		# For unique start-points:
		mins1 = pib_minusps (issqa1, issqa2, issqc1, issqc2, issqg1, issqg2, issqt1, issqt2, th3)
		# print 'mins = ', mins
		
		# For coverages:
		min2 = pib_minreads (icma1, icma2, icmc1, icmc2, icmg1, icmg2, icmt1, icmt2, th1)
		# print 'min = ', min
		
		# For unique start-points:
		mins2 = pib_minusps (issma1, issma2, issmc1, issmc2, issmg1, issmg2, issmt1, issmt2, th3)
		# print 'mins = ', mins
		
		# For coverages:
		min3 = pib_minreads (icmma1, icmma2, icmmc1, icmmc2, icmmg1, icmmg2, icmmt1, icmmt2, th1)
		# print 'min = ', min
		
		# For unique start-points:
		mins3 = pib_minusps (issmma1, issmma2, issmmc1, issmmc2, issmmg1, issmmg2, issmmt1, issmmt2, th3)
		# print 'mins = ', mins
		
		# For coverages:
		min4 = pib_minreads (icmqa1, icmqa2, icmqc1, icmqc2, icmqg1, icmqg2, icmqt1, icmqt2, th1)
		# print 'min = ', min
		
		# For unique start-points:
		mins4 = pib_minusps (issmqa1, issmqa2, issmqc1, issmqc2, issmqg1, issmqg2, issmqt1, issmqt2, th3)
		# print 'mins = ', mins
		
		#######################################################################################
		# compute consensus genotype and sum of reads per allele for each case
		#######################################################################################

		# For RQVmin, MMmax, MLmin, QVmin thresholds : unique start points:
		# print 'issmqa1, issmqa2, issmqc1, issmqc2, issmqg1, issmqg2, issmqt1, issmqt2, min ', issmqa1, issmqa2, issmqc1, issmqc2, issmqg1, issmqg2, issmqt1, issmqt2, min
		cgsmq, ngsmq = pib_genotype (issmqa1, issmqa2, issmqc1, issmqc2, issmqg1, issmqg2, issmqt1, issmqt2, mins4)

		# For RQVmin, MMmax, MLmin, QVmin thresholds : coverage:
		cgcmq, ngcmq = pib_genotype (icmqa1, icmqa2, icmqc1, icmqc2, icmqg1, icmqg2, icmqt1, icmqt2, min4)


		# For MMmax, MLmin, QVmin thresholds : unique start points:
		# print 'issmma1, issmma2, issmmc1, issmmc2, issmmg1, issmmg2, issmmt1, issmmt2, min ', issmma1, issmma2, issmmc1, issmmc2, issmmg1, issmmg2, issmmt1, issmmt2, min
		cgsmm, ngsmm = pib_genotype (issmma1, issmma2, issmmc1, issmmc2, issmmg1, issmmg2, issmmt1, issmmt2, mins3)

		# For MMmax, MLmin, QVmin thresholds : coverage:
		cgcmm, ngcmm = pib_genotype (icmma1, icmma2, icmmc1, icmmc2, icmmg1, icmmg2, icmmt1, icmmt2, min3)


		# For MLmin, QVmin thresholds : unique start points:
		cgsm, ngsm = pib_genotype (issma1, issma2, issmc1, issmc2, issmg1, issmg2, issmt1, issmt2, mins2)

		# For MLmin, QVmin thresholds : coverage:
		cgcm, ngcm = pib_genotype (icma1, icma2, icmc1, icmc2, icmg1, icmg2, icmt1, icmt2, min2)


		# For QVmin threshold : unique start points:
		cgsq, ngsq = pib_genotype (issqa1, issqa2, issqc1, issqc2, issqg1, issqg2, issqt1, issqt2, mins1)

		# For QVmin threshold : coverage:
		cgcq, ngcq = pib_genotype (icqa1, icqa2, icqc1, icqc2, icqg1, icqg2, icqt1, icqt2, min1)


		# For no filtering : unique start points:
		cgs, ngs = pib_genotype (issa1, issa2, issc1, issc2, issg1, issg2, isst1, isst2, mins)

		# For no filtering : coverage:
		cgc, ngc = pib_genotype (ica1, ica2, icc1, icc2, icg1, icg2, ict1, ict2, min)
		
		#######################################################################################
		# Decide on best genotype and its "confidence" tag:
		#######################################################################################
	
		ncovs = pib_nreads  (ica1, ica2, icc1, icc2, icg1, icg2, ict1, ict2)
		nusps = pib_nreads  (issa1, issa2, issc1, issc2, issg1, issg2, isst1, isst2)
		
		ncovs4= pib_nreads (icmqa1, icmqa2, icmqc1, icmqc2, icmqg1, icmqg2, icmqt1, icmqt2)
		nusps4= pib_nreads (issmqa1, issmqa2, issmqc1, issmqc2, issmqg1, issmqg2, issmqt1, issmqt2)
		#print ' ncovs, nusps, ncovs4, nusps4: ', ncovs, nusps, ncovs4, nusps4
		
		cgbest = pib_identical_genotype (cgsmq, cgcmq, cgsmm, cgcmm, cgsm, cgcm, cgsq, cgcq, cgs, cgc, ncovs, nusps, ncovs4, nusps4, cla, th2, th4)
		cgbestconf = ''

		# compute whether the coordinate lies in a homologous region or in a hypervariable region
		homolog, hypervar = pib_homolog_hypervar (issmqa1, 
			issmqa2, issmqc1, issmqc2, issmqg1, issmqg2, issmqt1, issmqt2,
			icmqa1, icmqa2, icmqc1, icmqc2, icmqg1, icmqg2, icmqt1, icmqt2,
			issmma1, issmma2, issmmc1, issmmc2, issmmg1, issmmg2, issmmt1, issmmt2,
			icmma1, icmma2, icmmc1, icmmc2, icmmg1, icmmg2, icmmt1, icmmt2,
			issma1, issma2, issmc1, issmc2, issmg1, issmg2, issmt1, issmt2,
			icma1, icma2, icmc1, icmc2, icmg1, icmg2, icmt1, icmt2,
			issqa1, issqa2, issqc1, issqc2, issqg1, issqg2, issqt1, issqt2,
			icqa1, icqa2, icqc1, icqc2, icqg1, icqg2, icqt1, icqt2,
			issa1, issa2, issc1, issc2, issg1, issg2, isst1, isst2, 
			ica1, ica2, icc1, icc2, icg1, icg2, ict1, ict2, th6, th7)
		
		if cgbest == '?':
			cgbest, cgbestconf = pib_best_genotype (cgsmq, cgcmq, 
			cgsmm, cgcmm, cgsm, cgcm, cgsq, cgcq, cgs, cgc,
			issmqa1, issmqa2, issmqc1, issmqc2, issmqg1, issmqg2, issmqt1, issmqt2,
			icmqa1, icmqa2, icmqc1, icmqc2, icmqg1, icmqg2, icmqt1, icmqt2,
			issmma1, issmma2, issmmc1, issmmc2, issmmg1, issmmg2, issmmt1, issmmt2,
			icmma1, icmma2, icmmc1, icmmc2, icmmg1, icmmg2, icmmt1, icmmt2,
			issma1, issma2, issmc1, issmc2, issmg1, issmg2, issmt1, issmt2,
			icma1, icma2, icmc1, icmc2, icmg1, icmg2, icmt1, icmt2,
			issqa1, issqa2, issqc1, issqc2, issqg1, issqg2, issqt1, issqt2,
			icqa1, icqa2, icqc1, icqc2, icqg1, icqg2, icqt1, icqt2,
			issa1, issa2, issc1, issc2, issg1, issg2, isst1, isst2, 
			ica1, ica2, icc1, icc2, icg1, icg2, ict1, ict2, cla, th2, th4, 
			homolog, hypervar, oldver, version )

		#print 'cgbest: ', cgbest
			
		#######################################################################################
		# Check whether alleles A and B of best genotype confirmed in forward and reverse read directions
		#######################################################################################
		
		# threshold: read direction with least read counts must have at least 20% of counts of reads of other read direction.
		# e.g.: forward: 5 counts, reverse 1 count --> confirmed allele 
		
		aconfirm=''
		bconfirm=''
		if cgbest != '' :
			try:
				aconfirm, bconfirm = pib_forward_reverse_confirm (cgbest,
				issmqa1, issmqa2, issmqc1, issmqc2, issmqg1, issmqg2, issmqt1, issmqt2,
				icmqa1, icmqa2, icmqc1, icmqc2, icmqg1, icmqg2, icmqt1, icmqt2,
				issmma1, issmma2, issmmc1, issmmc2, issmmg1, issmmg2, issmmt1, issmmt2,
				icmma1, icmma2, icmmc1, icmmc2, icmmg1, icmmg2, icmmt1, icmmt2,
				issma1, issma2, issmc1, issmc2, issmg1, issmg2, issmt1, issmt2,
				icma1, icma2, icmc1, icmc2, icmg1, icmg2, icmt1, icmt2,
				issqa1, issqa2, issqc1, issqc2, issqg1, issqg2, issqt1, issqt2,
				icqa1, icqa2, icqc1, icqc2, icqg1, icqg2, icqt1, icqt2,
				issa1, issa2, issc1, issc2, issg1, issg2, isst1, isst2, 
				ica1, ica2, icc1, icc2, icg1, icg2, ict1, ict2, cla, th5 )
			except:
				print '*** In line: \n', line
				print '*** Outputting this genotype line without read direction info and continuing with next genotype line'
		
		#######################################################################################
		# write to tab-separated output file:
		#######################################################################################

		#c='%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t\n' % \
		#(ichrom+1, poi+1, refbasepoi, ctxt, gcc, cla, indelread, issqa1,issqa2,issqc1,issqc2,issqg1,issqg2,issqt1,issqt2,issqn1,issqn2, icqa1,icqa2,icqc1,icqc2,icqg1,icqg2,icqt1,icqt2,icqn1,icqn2, issa1,issa2,issc1,issc2,issg1,issg2,isst1,isst2,issn1,issn2, ica1,ica2,icc1,icc2,icg1,icg2,ict1,ict2,icn1,icn2)

		c='%s\t%s\t%s\t%s\t%s\t%s\t%s' % (ichrom, poi, refbasepoi, ctxt, gcc, cla, indelread)

		# Best genotype and its "confidence" class:
		c+='\t%s\t%s' % (cgbest,cgbestconf)

		# Allele confirmation by sufficient forward and reverse reads:
		c+='\t%s\t%s' % (aconfirm,bconfirm)

		# Consensus Genotype:
		c+='\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s' % \
		(cgsmq,cgcmq, cgsmm,cgcmm, cgsm,cgcm, cgsq,cgcq, cgs,cgc)

		# Reads:
		c+='\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s' % \
		(ngsmq,ngcmq, ngsmm,ngcmm, ngsm,ngcm, ngsq,ngcq, ngs,ngc)


		# Homologous, Hypervariable:
		if version != '1.4.3':
			c+='\t%s\t%s' % (homolog,hypervar)

			
		# For RQVmin, MMmax, MLmin, QVmin thresholds : unique start points:
		c+='\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s' % \
		(issmqa1, issmqa2,  issmqc1, issmqc2,  issmqg1, issmqg2,  issmqt1, issmqt2,  issmqn1, issmqn2)

		# For RQVmin, MMmax, MLmin, QVmin thresholds : coverage:
		c+='\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s' % \
		(icmqa1, icmqa2,  icmqc1, icmqc2,  icmqg1, icmqg2,  icmqt1, icmqt2,  icmqn1, icmqn2)


		# For MMmax, MLmin, QVmin thresholds : unique start points:
		c+='\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s' % \
		(issmma1, issmma2,  issmmc1, issmmc2,  issmmg1, issmmg2,  issmmt1, issmmt2,  issmmn1, issmmn2)
		
		# For MMmax, MLmin, QVmin thresholds : coverage:
		c+='\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s' % \
		(icmma1, icmma2,  icmmc1, icmmc2,  icmmg1, icmmg2,  icmmt1, icmmt2,  icmmn1, icmmn2)


		# For MLmin, QVmin thresholds : unique start points:
		c+='\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s' % \
		(issma1, issma2,  issmc1, issmc2,  issmg1, issmg2,  issmt1, issmt2,  issmn1, issmn2)

		# For MLmin, QVmin thresholds : coverage:
		c+='\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s' % \
		(icma1, icma2,  icmc1, icmc2,  icmg1, icmg2,  icmt1, icmt2,  icmn1, icmn2)


		# For QVmin threshold : unique start points:
		c+='\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s' % \
		(issqa1, issqa2,  issqc1, issqc2,  issqg1, issqg2,  issqt1, issqt2,  issqn1, issqn2)

		# For QVmin threshold : coverage:
		c+='\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s' % \
		(icqa1, icqa2,  icqc1, icqc2,  icqg1, icqg2,  icqt1, icqt2,  icqn1, icqn2)


		# For no filtering : unique start points:
		c+='\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s' % \
		(issa1, issa2,  issc1, issc2,  issg1, issg2,  isst1, isst2,  issn1, issn2)

		# For no filtering : coverage:
		c+='\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s' % \
		(ica1, ica2,  icc1, icc2,  icg1, icg2,  ict1, ict2,  icn1, icn2)

		# End of line:
		c+='\n'
		out.write (c)

	
	#######################################################################################################
	# End of WHILE loop over all positions of interest
	#######################################################################################################

	out.close()
	for i in inf:
		i.close()

###########################################################################################################
#
# the actual main
#
###########################################################################################################

# print 'step 1'

###########################################################################################################
# set resource limits:
###########################################################################################################

setlimits()
# print 'step 2'

###########################################################################################################
# get command line arguments
###########################################################################################################

nargs = len(sys.argv)
# print nargs
# print 'step 3'
if nargs < 3:
	print pib_consensus.__doc__
	sys.exit()

# Process the optional arguments:
# pibase_consensus [-t[h][v [ver]] t1 t2 t3 t4 t5 {h1 h2}] in1 [in2 ... inN] out
#  0                 1              2  3  4  5  6           7                  8
#  0                   1            2  3  4  5  6   7  8    9                 10
#  0                      1   2     3  4  5  6  7           8                  9  (version 1.4.3)
#  0                      1   2     3  4  5  6  7   8  9   10                 11  (> version 1.4.3)
optargs = ['-t', '-th', '-tv', '-thv']
version='1.4.5'
oldver=0
h=0
if sys.argv[1] in optargs:
	# There are user-specified thresholds for allele counts
	
	if optargs.index(sys.argv[1]) > 1:
		# revert to an available old version
		oldversions = ['1.4.3']
		if nargs < 9 or sys.argv[2] not in oldversions:
			print pib_consensus.__doc__
			print '        # Available old versions that can be reverted to:'
			for i in oldversions:
				print '        # %s' % (i)
			print '        #######################################################################################'
			sys.exit()
		elif sys.argv[1]=='-thv' and sys.argv[2]==oldversions[0]:
			print pib_consensus.__doc__
			print '*** ERROR: Option "-thv" not available in old version %s \n' % (oldversions[0])
			sys.exit()
			
		else:
			version = sys.argv[2]
			oldver=1
			# start of infilenames in list sys.argv:
			if version==oldversions[0] or sys.argv[1]=='-t' or sys.argv[1]=='-tv':
				k=8
			else:
				h=1
				k=10
		# start of thresholds th1, th2, etc. in list sys.argv: 
		j=3
	else:
		# No reversion request. Only user-specified thresholds for allele counts
		j=2
		if  sys.argv[1]=='-th':
			h=1
			k=9
		else:
			k=7
		if nargs <= k:
			print pib_consensus.__doc__
			sys.exit()
	# print ' j: %s' % (j)
	try:
		# allele thresholds for SNP-calling
		th1 = float(sys.argv[j])
		th2 = int(sys.argv[j+1])
		th3 = float(sys.argv[j+2])
		th4 = int(sys.argv[j+3])
		# threshold for both-stranded validation
		th5 = float(sys.argv[j+4])
	except:
		print pib_consensus.__doc__
		print '*** ERROR in arguments t1 t2 t3 t4 t5: %s %s %s %s %s \n' % (sys.argv[j], 
		sys.argv[j+1], sys.argv[j+2], sys.argv[j+3], sys.argv[j+4])
		sys.exit()
	try:
		# thresholds for homologous regions and hypervariable regions
		if h==1:
			h1 = float(sys.argv[j+5])
			h2 = int(sys.argv[j+6])
		else:
			h1 = 0.022
			h2 = 2
	except:
		print pib_consensus.__doc__
		print '*** ERROR in arguments h1 h2: %s %s \n' % (sys.argv[j+5], sys.argv[j+6])
		sys.exit()
else:
	# There are no user-specified thresholds for allele counts
	# Use defaults:
	k = 1
	th1 = 0.022
	th2 = 8
	th3 = 0.04
	th4 = 4
	th5 = 0.2
	h1 = 0.022
	h2 = 2

argu = sys.argv [k-1:]
nargu = nargs - k + 1

outfile = sys.argv[nargs-1]
# print 'outfile = '+outfile

###########################################################################################################
# call main procedure
###########################################################################################################

pib_consensus (argu, nargu, outfile, th1, th2, th3, th4, th5, h1, h2, oldver, version)

print('*** Completed successfully')


