#!/usr/bin/env python
##!/usr/bin/python

import sys
import resource
import time

__author__="mforster"
__date__ ="$Feb 22, 2011 15:05 PM$"

def setlimits():
	#############################################################################
	# sets resource limits (mandatory for job interoperability on cluster)
	# to 10MB Stack + 1GB data + 1GB RLIMIT_RSS
	# first steps taken from
	# http://docs.python.org/library/resource.html and
	# http://www.doughellmann.com/PyMOTW/resource/
	# Author: MICHAEL FORSTER m.forster@ikmb.uni-kiel.de
	# Version: 13 September 2010
	#############################################################################

	#############################################################################
	# Adjust the maximum size (in bytes) of the process's heap
	#############################################################################

	soft, hard = resource.getrlimit(resource.RLIMIT_DATA)
	#print 'Data/heap limit starts as  :', soft

	# 1GB heap:
	newheap=1024*1024*1024

	resource.setrlimit(resource.RLIMIT_DATA, (newheap, hard))

	soft, hard = resource.getrlimit(resource.RLIMIT_DATA)
	#print 'Data/heap limit changed to :', soft

	#############################################################################
	# ca 10MB original stack changed to exactly 10MB stack:
	#############################################################################

	#newstack=10*1024*1024+1
	newstack=10*1024*1024
	soft, hard = resource.getrlimit(resource.RLIMIT_STACK)
	#print 'Stack limit starts as  :', soft
	resource.setrlimit(resource.RLIMIT_STACK, (newstack, hard))
	soft, hard = resource.getrlimit(resource.RLIMIT_STACK)
	#print 'Stack limit changed to :', soft

	#############################################################################
	# 1GB RLIMIT_RSS
	# "Maximale resident set size in Byte. Bei Speicherengpaessen entzieht der Kern
	# den Prozessen, die den resident set size ueberschreiten, den zuviel
	# angeforderten Speicher."
	# http://pronix.linuxdelta.de/C/Linuxprogrammierung/Linuxsystemprogrammieren_C_Kurs_Kapitel3a.shtml
	#############################################################################

	newrss=1024*1024*1024
	soft, hard = resource.getrlimit(resource.RLIMIT_RSS)
	#print 'RLIMIT_RSS starts as  :', soft
	resource.setrlimit(resource.RLIMIT_RSS, (newrss, hard))
	soft, hard = resource.getrlimit(resource.RLIMIT_RSS)
	#print 'RLIMIT_RSS changed to :', soft

def pib_diff (infile1, infile2, thr, outfile):
	'''        #######################################################################################
	#  pibase_diff v1.4.3  -  comparison of two samples by genotype (not by fisherdiff!). 
	#######################################################################################
	#  pibase_diff  <control>  <case>  <outfile>  <threshold>
	#  
	#  <control>	pibase_consensus genotype file of control (normal) sample
	#  <case>	pibase_consensus genotype file of case (diseased) sample
	#  <threshold>	maximal genotype "BestQual" level to consider (0 to 8; 8 = worst)
	#  <outfile>	output file. Shared genotypes denoted by "=", 
	# 		genotypes only in control sample by "-", 
	#		genotypes only in disease sample by "+"
	#######################################################################################
	#  Author: Michael Forster, m.forster@ikmb.uni-kiel.de         
	#  Date: 22 February 2011, 25 July 2011
	#######################################################################################
	'''
	
	print 'control file: ', infile1
	print 'case file   : ', infile2
	print 'outfile: ', outfile
	
	#######################################################################################
	# build dictionary of snp positions and variants for infile 2:
	#######################################################################################
	
	inf1 = open(infile1, 'r')
	inf2 = open(infile2, 'r')
	
	# find table header
	while True :
		line = inf2.readline()
		if len(line) == 0:
			# EOF
			sys.exit (' *** Error: End of file %s reached before table header could be found' % (infile2))
		if line.startswith('# C\tPos[1]\tRef\tRefSeqContext\tGC[%]\tClass\tIgn\tBestGen\tBestQual'):
			header = line
			break
	
	# build dictionary
	cancer={}
	while True :
		line = inf2.readline()
		if len(line) == 0:
			# EOF
			break
		cols = line.split('\t')
		
		pos = cols[0] + '_' + cols[1]
		#chrom = int(cols[0])
		#pos1 = int(cols[1])
		#ref = cols[2]
		bestgen = cols[7]
		bestconf = cols[8]
		# reads = cols[28]
		# cancer[pos] = (chrom, pos1, ref, bestgen, bestconf, reads)
		cancer[pos] = (bestgen, bestconf, line)

	# print cancer
	
	# sys.exit()
	
	#######################################################################################
	# Open outfile and write header lines
	#######################################################################################
	
	out=open(outfile,'w')
	outtime = time.strftime("%Y_%m_%d__%H:%M:%S") + "\n"
	out.write ('# File created by pibase_diff v1.4.3 on '+ outtime)
	out.write ('# Control sample file: '+infile1+'\n')
	out.write ('# Case sample file: '+infile2+'\n#\n')
	out.write ('# Shared genotypes denoted by "=", genotypes only in control sample by "-", '+
	'genotypes only in case sample by "+"\n#\n')
	headerout = '#D\t' + header[2:]
	out.write (headerout)
	
	#######################################################################################
	# Read infile1 line by line and
	#######################################################################################
		
	# find table header
	while True :
		line = inf1.readline()
		if len(line) == 0:
			# EOF
			sys.exit (' *** Error: End of file %s reached before table header could be found' % (infile1))
		if line.startswith('# C\tPos[1]\tRef\tRefSeqContext\tGC[%]\tClass\tIgn\tBestGen\tBestQual'):
			break
			
	while True :
		line = inf1.readline()

		if len(line) == 0:
			# EOF
			break
		cols = line.split('\t')
		
		pos = cols[0] + '_' + cols[1]
		#chrom = int(cols[0])
		#pos1 = int(cols[1])
		#ref = cols[2]
		bestgen = cols[7]
		bestconf = cols[8]
		#reads = cols[28]
		
		# get genotype confidence for the control sample:
		if bestconf == '':
			bestc = 0
		else:
			bestc = int(bestconf[1:])
		
		try:
			# get genotype and confidence for the disease sample:
			#chrom2, pos12, ref2, bestgen2, bestconf2, reads2 = cancer[pos]
			bestgen2, bestconf2, line2 = cancer[pos]
			if bestconf2 == '':
				bestc2 = 0
			else:
				bestc2 = int(bestconf2[1:])
			#print 'bestc, bestc2', bestc, bestc2
			# consider this snp only if acceptable genotype confidence in both samples:
			if bestc2 <= thr and bestc <= thr :
				if bestgen == bestgen2:
					# genotypes are identical, take the worst-confidence sample
					if bestc2 > bestc:
						c = '=\t' + line2
					else:
						c = '=\t' + line
					out.write(c)
				else:
					# normal sample genotype:
					c = '-\t' + line
					out.write(c)
					# disease sample genotype:
					c = '+\t' + line2
					out.write(c)
		except:
			# this snp does not occur in infile2
			# (could be due to low/no coverage)
			# so "comment out" this position
			c = '##\t' + line
			out.write(c)

	out.close()
	inf1.close()
	inf2.close()

##############################################
# the actual main
##############################################

if len(sys.argv) < 5:
	print pib_diff.__doc__
	sys.exit()
else:
	infile1 = sys.argv[1]
	infile2 = sys.argv[2]
	thr = int(sys.argv[3])
	outfile = sys.argv[4]

setlimits()

pib_diff (infile1, infile2, thr, outfile)
print('Done')
