#!/usr/bin/env python
##!/usr/bin/python

import sys
import resource
import time
from subprocess import call
import os

__author__="mforster"
__date__ ="$Mar 23, 2011 11:00 AM$"

def setlimits():
	#############################################################################
	# sets resource limits (mandatory for job interoperability on cluster)
	# to 10MB Stack + 1GB data + 1GB RLIMIT_RSS
	# first steps taken from
	# http://docs.python.org/library/resource.html and
	# http://www.doughellmann.com/PyMOTW/resource/
	# Author: MICHAEL FORSTER m.forster@ikmb.uni-kiel.de
	# Version: 13 September 2010
	#############################################################################

	#############################################################################
	# Adjust the maximum size (in bytes) of the process's heap
	#############################################################################

	soft, hard = resource.getrlimit(resource.RLIMIT_DATA)
	#print 'Data/heap limit starts as  :', soft

	# 1GB heap:
	newheap=1024*1024*1024

	resource.setrlimit(resource.RLIMIT_DATA, (newheap, hard))

	soft, hard = resource.getrlimit(resource.RLIMIT_DATA)
	#print 'Data/heap limit changed to :', soft

	#############################################################################
	# ca 10MB original stack changed to exactly 10MB stack:
	#############################################################################

	#newstack=10*1024*1024+1
	newstack=10*1024*1024
	soft, hard = resource.getrlimit(resource.RLIMIT_STACK)
	#print 'Stack limit starts as  :', soft
	resource.setrlimit(resource.RLIMIT_STACK, (newstack, hard))
	soft, hard = resource.getrlimit(resource.RLIMIT_STACK)
	#print 'Stack limit changed to :', soft

	#############################################################################
	# 1GB RLIMIT_RSS
	# "Maximale resident set size in Byte. Bei Speicherengpaessen entzieht der Kern
	# den Prozessen, die den resident set size ueberschreiten, den zuviel
	# angeforderten Speicher."
	# http://pronix.linuxdelta.de/C/Linuxprogrammierung/Linuxsystemprogrammieren_C_Kurs_Kapitel3a.shtml
	#############################################################################

	newrss=1024*1024*1024
	soft, hard = resource.getrlimit(resource.RLIMIT_RSS)
	#print 'RLIMIT_RSS starts as  :', soft
	resource.setrlimit(resource.RLIMIT_RSS, (newrss, hard))
	soft, hard = resource.getrlimit(resource.RLIMIT_RSS)
	#print 'RLIMIT_RSS changed to :', soft

def pib_fd_cov_ab (line, ksp4):
	#############################################################################
	# split line, get column ksp4+1 with raw coverages, find major
	# and minor alleles, and compute raw coverage
	#############################################################################
	cols = line.split('\t')
	covs = cols[ksp4 + 1].split(':')
	# get allelic coverages: and total coverage
	cov = []
	c = 0
	for i in range(0,4):
		j = int(covs[i])
		cov.append(j)
		c += j
	# find major allele
	amax = -1
	for i in range(0,4):
		if cov[i] > amax:
			ia = i
			amax = cov[i]
	# find minor allele
	bmax = -1
	for i in range(0,4):
		if i == ia:
			continue
		elif cov[i] > bmax:
			ib = i
			bmax = cov[i]
	return ia, ib, amax, bmax, c

def pib_fisherdiff (infile1, infile2, outfile, cov, pthres, fac):
	'''        #######################################################################################
	#  pibase_fisherdiff v1.4.3  -  fisher's 2x4 exact test for
	#  comparing two samples (e.g. healthy control sample and diseased case sample).
	#######################################################################################
	#  pibase_fisherdiff  <control>  <case>  <outfile> [[cov] [[p] [[fac]]]
	#  
	#  <control>  pibase_consensus genotype file of control sample
	#  <case>     pibase_consensus genotype file of diseased sample
	#  <outfile>  output file. Shared genotype pairs denoted by "=-" and "=+", 
	#             genotypes only in control sample by "-", 
	#             genotypes only in disease sample by "+"
	#             low coverage in one or both samples by "?-" and "?+"
	#             Using two-tailed p-value for fisher's exact test on ACGT to decide
	#             whether genotypes are shared or different
	#  [cov]      min coverage threshold (optional. Default: 100)
	#  [p]        max median p-value threshold (optional. Default: 0.01)
	#  [fac]      max factor (optional. Default: 100). 
	#             If major alleles are identical in case and control, 
	#             major allele counts / minor allele counts must be <= fac
	#######################################################################################
	#  Author: Michael Forster, m.forster@ikmb.uni-kiel.de
	#  Date:   23 March 2011, 17 May 2011
	#######################################################################################
	'''
	
	print 'control (i.e. non-disease) file: ', infile1
	print 'case sample (i.e. disease) file: ', infile2
	print 'outfile:                         ', outfile
	print 'options: coverage >= %s, p.med <= %s, major_allele_counts / minor_allele_counts <= %s' % (cov, pthres, fac)
	
	#######################################################################################
	# Open input files
	# read all lines until table headers, 
	# find columns of start-point counts
	#######################################################################################
	
	inf1 = open(infile1, 'r')
	inf2 = open(infile2, 'r')
	
	# find table header of file 1
	while True :
		line = inf1.readline()
		if len(line) == 0:
			# EOF
			sys.exit (' *** Error: End of file %s reached before table header could be found' % (infile1))
		if line.startswith('# C\tPos[1]\tRef\tRefSeqContext\tGC[%]\tClass\tIgn\tBestGen\tBestQual'):
			break
	# find table header of file 2
	while True :
		line = inf2.readline()
		if len(line) == 0:
			# EOF
			sys.exit (' *** Error: End of file %s reached before table header could be found' % (infile2))
		if line.startswith('# C\tPos[1]\tRef\tRefSeqContext\tGC[%]\tClass\tIgn\tBestGen\tBestQual'):
			header = line
			break
	
	# find columns of start-point-counts
	cols = line.split('\t')
	ksp=[]
	k=-1
	for col in cols:
		k += 1
		if col == 'SP':
			# this is the genotype column. The first counts column is 10 tabs further.
			for j in range(0,5):
				ksp.append (k + 10 + j*2)
			break
	if len(ksp) == 0:
		sys.exit (' *** Error: In file %s table header does not contain "SP"' % (infile2))
	# print 'SP columns (0-based): ksp: ',ksp
		
	#######################################################################################
	# Open outfile and write header lines
	#######################################################################################
	
	out=open(outfile,'w')
	datstr = time.strftime("%Y_%m_%d__%H_%M_%S")
	outtime = datstr + "\n"
	out.write ('# File created by pibase_fisherdiff v1.4.3 on '+ outtime)
	out.write ('# Control sample file: '+infile1+'\n')
	out.write ('# Disease sample file: '+infile2+'\n#\n')
	out.write ('# Shared genotypes denoted by "=-" and "=+",' +
	' genotypes only in control sample by "-", genotypes only in disease sample by "+"\n')
	out.write ('# Shared genotypes defined by: \n')
	out.write ('#  p.med >= %s AND coverage >= %s AND major allele counts / minor allele counts <= %s\n'  % (pthres, cov, fac))
	out.write ('#  where p.med is the median of p.f4 to p.f0, and where f4 to f0 denote the filter levels (filter settings see control and disease sample files)\n')
	out.write ("#  p is the two-tailed p-value for Fisher's exact test for the 2x4 matrix of the 4 unique start point counts (ACGT) for the 2 samples (control, disease/case).\n")
	out.write ("#  If the coverage falls below the threshold, D_error is output with an '!' and bit 0 (control) or 1 (case) is set.\n")
	out.write ("#  If the major/minor allele counts threshold is exceeded, D_error is output with an '!' and bit 2 (control) or 3 (case) is set.\n#\n")
	
	headerout = '#D\tD_error\tp.med\tp.f4\tp.f3\tp.f2\tp.f1\tp.f0\t' + header[2:]
	out.write (headerout)
	
	#######################################################################################
	# Open file for pibase_fisherdiff_643 and write header lines
	#######################################################################################
	
	datstr = time.strftime("%Y_%m_%d__%H_%M_%S")
	tmp_in = 'tmp_in_pibase_fisherdiff_643__'+datstr+'.dat'
	tmp_out = 'tmp_out_pibase_fisherdiff_643__'+datstr+'.out'
	tmpi = open (tmp_in,'w')
	
	tmpi.write('version=1\n')
	tmpi.write(infile1 + '\n')
	tmpi.write(infile2 + '\n')

	#######################################################################################
	# Read line by line from both input files and 
	# write start-point-counts to tmp file for pibase_fisherdiff_643
	#######################################################################################
	
	while True :
		# read cols from file 1:
		line1 = inf1.readline()
		if len(line1) == 0:
			# EOF
			break
		cols1 = line1.split('\t')
		pos1 = cols1[0] + '_' + cols1[1]
		
		# read cols from file 2:
		line2 = inf2.readline()
		if len(line2) == 0:
			# EOF
			break
		cols2 = line2.split('\t')
		pos2 = cols2[0] + '_' + cols2[1]

		# check that both lines refer to the same chr and position:
		if pos1 != pos2:
			sys.exit (' *** Error: In input file %s position chr%s does not\n *** correspond to input file %s position chr%s' % (infile1, line1, pos1, infile2, line2, pos2))
		
		# write position
		tmpi.write( cols2[0] + ' ' + cols2[1] + '\n')
		
		# write five 2x4 tables of allele counts for ACGT (i.e. for filter levels 0,1,2,3,4):
		for i in range(0,5):
			spc = cols1 [ ksp[i] ]
			spcn = spc.replace (':',' ')
			tmpi.write (spcn + '\n')
			spc = cols2 [ ksp[i] ]
			spcn = spc.replace (':',' ')
			tmpi.write (spcn + '\n')
	tmpi.close()
	inf1.close()
	inf2.close()
	
	#######################################################################################
	# run pibase_fisherdiff_643
	#######################################################################################
	
	retcode = call("pibase_fisherdiff_643 " + tmp_in + " " + tmp_out, shell=True)
	# print 'retcode :', retcode
	if retcode < 0:
		sys.exit (' *** Error: retcode %s in pibase_fisherdiff_643\n *** files: %s and %s' % (retcode, tmp_in, tmp_out))
	
	# clean up tmp_in
	retcode = call("rm -f " + tmp_in, shell=True)
	
	#######################################################################################
	# merge p-values and pibase_consensus lines into outfile
	#######################################################################################
	
	inf1 = open(infile1, 'r')
	inf2 = open(infile2, 'r')
	tmpo = open(tmp_out,'r')
		
	# find table header of file 1
	while True :
		line = inf1.readline()
		if len(line) == 0:
			# EOF
			sys.exit (' *** Error: End of file %s reached before table header could be found' % (infile1))
		if line.startswith('# C\tPos[1]\tRef\tRefSeqContext\tGC[%]\tClass\tIgn\tBestGen\tBestQual'):
			break
	# find table header of file 2
	while True :
		line = inf2.readline()
		if len(line) == 0:
			# EOF
			sys.exit (' *** Error: End of file %s reached before table header could be found' % (infile2))
		if line.startswith('# C\tPos[1]\tRef\tRefSeqContext\tGC[%]\tClass\tIgn\tBestGen\tBestQual'):
			header = line
			break
	# check header of tmp_out:
	line = tmpo.readline().strip()
	if line != 'version=1':
		sys.exit (' *** Error: File %s line 1 is %s, must be "version=1"' % (tmp_out,line))
	line = tmpo.readline().strip()
	if line != infile1:
		sys.exit (' *** Error: File %s line 2 is %s, must be %s' % (tmp_out,line,infile1))
	line = tmpo.readline().strip()
	if line != infile2:
		sys.exit (' *** Error: File %s line 3 is %s, must be %s' % (tmp_out,line,infile2))
		
	while True :
		# read chr#, pos, median p-value, and p-values:
		line = tmpo.readline().strip()
		if len(line) == 0:
			# EOF
			break
		cols = line.split()
		poso = cols[0]+'_'+cols[1]
		
		line1 = inf1.readline()
		if len(line1) == 0:
			# EOF
			sys.exit (' *** Error: Premature EOF reached in %s for position %s' % (infile1, poso))
		# split line, get column ksp4+1 with raw coverages, find major
		# and minor alleles, and compute raw coverage
		ia1, ib1, a1, b1, c1 = pib_fd_cov_ab (line1, ksp[4])
		
		line2 = inf2.readline()
		if len(line2) == 0:
			# EOF
			sys.exit (' *** Error: Premature EOF reached in %s for position %s' % (infile2, poso))
		# split line, get column ksp4+1 with raw coverages, find major
		# and minor alleles, and compute raw coverage
		ia2, ib2, a2, b2, c2 = pib_fd_cov_ab (line2, ksp[4])
		
		# build error tag variable derr:
		derr = 0
		if c1 < cov:
			derr += 1
		if c2 < cov:
			derr += 2
		if ia1 == ia2:
			if b1 > 0:
				if a1/b1 > fac:
					derr += 4
			if b2 > 0:
				if a2/b2 > fac:
					derr += 8
		# build error tag straing D_err:
		if derr == 0:
			sderr = '\t'
		else:
			sderr = '!%s\t' % (derr)
		
		pmed = float(cols[2].strip())
		for i in range(2,8):
			cols[i]=cols[i].strip()+'\t'
		
		if pmed < 0 or (derr > 0 and pmed < pthres):
			s = '?'
		elif pmed < pthres :
			s = ''
		else:
			s = '='
		out.write(s + '-\t' + sderr + cols[2] + cols[3] + cols[4] + cols[5] + cols[6] + cols[7] + line1)
		out.write(s + '+\t' + sderr + cols[2] + cols[3] + cols[4] + cols[5] + cols[6] + cols[7] + line2)
	out.close()
	tmpo.close()
	inf1.close()
	inf2.close()

	# clean up tmp_out
	retcode = call("rm -f " + tmp_out, shell=True)

##############################################
# the actual main
##############################################

n = len(sys.argv) 
if n < 4:
	print pib_fisherdiff.__doc__
	sys.exit()
else:
	infile1 = sys.argv[1]
	infile2 = sys.argv[2]
	outfile = sys.argv[3]
	cov = 100
	pthres = 0.01
	fac = 100
	if n > 4:
		cov = int(sys.argv[4])
	if n > 5:
		pthres = float(sys.argv[5])
	if n > 6:
		fac = float(sys.argv[6])
	
setlimits()

pib_fisherdiff (infile1, infile2, outfile, cov, pthres, fac)
print('Done')
