#!/usr/bin/env python
##!/usr/bin/python

import sys
import time
import resource

__author__="mforster"
__date__ ="$Oct 14, 2012 13:00 PM$"

def setlimits():
	#############################################################################
	# sets resource limits (mandatory for job interoperability on cluster)
	# to 10MB Stack + 1GB data + 1GB RLIMIT_RSS
	# first steps taken from
	# http://docs.python.org/library/resource.html and
	# http://www.doughellmann.com/PyMOTW/resource/
	# Author: MICHAEL FORSTER m.forster@ikmb.uni-kiel.de
	# Version: 13 September 2010
	#############################################################################

	#############################################################################
	# Adjust the maximum size (in bytes) of the process's heap
	#############################################################################

	soft, hard = resource.getrlimit(resource.RLIMIT_DATA)
	#print 'Data/heap limit starts as  :', soft

	# 1GB heap:
	newheap=1024*1024*1024

	resource.setrlimit(resource.RLIMIT_DATA, (newheap, hard))

	soft, hard = resource.getrlimit(resource.RLIMIT_DATA)
	#print 'Data/heap limit changed to :', soft

	#############################################################################
	# ca 10MB original stack changed to exactly 10MB stack:
	#############################################################################

	#newstack=10*1024*1024+1
	newstack=10*1024*1024
	soft, hard = resource.getrlimit(resource.RLIMIT_STACK)
	#print 'Stack limit starts as  :', soft
	resource.setrlimit(resource.RLIMIT_STACK, (newstack, hard))
	soft, hard = resource.getrlimit(resource.RLIMIT_STACK)
	#print 'Stack limit changed to :', soft

	#############################################################################
	# 1GB RLIMIT_RSS
	# "Maximale resident set size in Byte. Bei Speicherengpaessen entzieht der Kern
	# den Prozessen, die den resident set size ueberschreiten, den zuviel
	# angeforderten Speicher."
	# http://pronix.linuxdelta.de/C/Linuxprogrammierung/Linuxsystemprogrammieren_C_Kurs_Kapitel3a.shtml
	#############################################################################

	newrss=1024*1024*1024
	soft, hard = resource.getrlimit(resource.RLIMIT_RSS)
	#print 'RLIMIT_RSS starts as  :', soft
	resource.setrlimit(resource.RLIMIT_RSS, (newrss, hard))
	soft, hard = resource.getrlimit(resource.RLIMIT_RSS)
	#print 'RLIMIT_RSS changed to :', soft

#def pib_flgsnp_cov_ab (covf0, bestgen, ref, covmin, fmin):
def pib_flgsnp_cov_ab (covf0, bestgen, ref):
	#############################################################################
	# compute whether this is a snp, snp-allele-coverage, total coverage
	#############################################################################
	
	covs = covf0.split(':')
	#print 'covs',covs
	
	# convert Ref nucleotide to upper case (in case it is in a repeat region)
	#print ref
	refu = ref.upper()
	dict = {}
	dict['A'] = 0
	dict['C'] = 1
	dict['G'] = 2
	dict['T'] = 3
	#print 'refu, dict', refu, dict 
	iref = dict[refu]
	
	# get non-Reference coverage and total coverage
	cn = 0
	c = 0
	for i in range(0,4):
		j = int(covs[i])
		if iref != i:
			cn += j
		c += j
	
	# decide whether this is a potential SNP by comparing BestGen and Ref:
	refgen = refu+refu
	isnp = 0
	if refgen != bestgen and len(bestgen) > 1:
		#print ' refgen=%s:bestgen=%s: ' % (refgen,bestgen) 
		isnp = 1	
	
	#if c > 0:
	#	f = float(cn) / float(c)
	#	# flag this a snp if sufficient coverage
	#	# and sufficiently high fraction of non-ref reads:
	#	if cn >= covmin and f >= fmin:
	#		isnp=1
	
	return isnp, cn, c

#def pib_flgsnp (infile, outfile, covmin, fmin):
def pib_flgsnp (infile, outfile):
	'''        #######################################################################################
	#  pibase_flagsnp v1.4.3
	#######################################################################################
	#  Flags a BestGen genotype as being a snp if different to reference.
	#  Three columns "IsSnp" (1/0 flag), "SnpReads" (non-Ref coverage), "CovAll" 
	#  (all reads indicating A, C, G, or T but not N) are inserted before the "BestGen"
	#  column.
	#
	#  Usage:  pibase_flagsnp- <in> <out>
	#  where
	#  in:     tab-separated file 
	#          (from pibase_diff, pibase_fisherdiff or pibase_consensus)
	#  out:    results file with merging of annotations and infile
	#
	#######################################################################################
	#  Author: Michael Forster,                            m.forster@ikmb.uni-kiel.de
	#  Date:   14-17 October 2011,  17.10.2011, 18.10.2011
	#######################################################################################
	'''
	
	print 'Infile : ', infile
	print 'Outfile: ', outfile
	#print 'Covmin : ', covmin
	#print 'Fmin   : ', fmin
	
	############################################################################
	# open infile and outfile and copy infile header lines to outfile:
	############################################################################
	
	inf = open (infile, 'r')
	out = open (outfile, 'w')
	
	outtime = time.strftime("%Y_%m_%d__%H:%M:%S") + "\n"
	out.write ('# File created by pibase_flagsnp v1.4.3 on '+ outtime)
	out.write ('# from: %s\n' % (infile))
	#out.write ('# using minimal non-reference allele threshold %s  and fraction threshold %s   (at filter level 0)\n' % (covmin, fmin))
	#out.write ('# to flag genotypes as potential SNPs.\n')
	out.write ('# using BestGen/Ref difference to flag genotypes as potential SNPs.\n')
	out.write ('# Column "IsSnp" is 1 if the genotype is a potential SNP.\n')
	out.write ('# Column "SnpReads" counts reads which indicate a non-reference allele (at filter level 0).\n')
	out.write ('# Column "CovAll" (coverage) counts reads which indicate any allele (A, C, G, or T, but not N) (at filter level 0).\n#\n')
	


	n = 0
	while True:
		n += 1
		line = inf.readline()
		if len(line) == 0:
			# EOF
			sys.exit (' *** Error: End of file %s reached after %s lines before table header could be found' % (infile, n))
		if line.find('RefSeqContext\tGC[%]')> -1:
			header = line
			break
		else:
			# add two tabs in the respective two header lines
			if line.find('MapQV>RQVmin and MM<=MMmax') != -1:
				line=line.replace('# ','# \t\t\t')
			if line.find('Unique start points (on +- strands)') != -1:
				line=line.replace('# ','# \t\t\t')
			# output header line
			out.write(line)
		
	############################################################################
	# table header line for outfile
	############################################################################
	cols = header.split('\t')
	ncols = len(cols)
	
	# find column with "BestGen":
	bgc = -1
	for c in cols:
		bgc += 1
		if c=="BestGen":
			break
	if bgc == -1:
		print '*** Error: No BestGen column in table header line of file ',infile 
		sys.exit()
	
	outheader=''
	for i in range(0,bgc):
		outheader += cols[i] + '\t'
	outheader += 'IsSnp\tSnpReads\tCovAll'
	for i in range(bgc,ncols):
		outheader += '\t'+cols[i]
	out.write(outheader)
	
	############################################################################
	#  Find Ref column in table header
	############################################################################

	refc = -1
	for c in cols:
		refc += 1
		if c=="Ref":
			break
	if refc == -1 or refc >= len(cols)-1:
		print '*** Error: No "Ref" column in table header line of file ',infile 
		sys.exit()
		
	############################################################################
	#  Find coverage column for filter level f0 in table header
	############################################################################

	cvc = -1
	for c in cols:
		cvc += 1
		if c=="CV":
			# this is the first (f4) CV, so adjust to f0 CV:
			cvc = cvc + 18
			break
	if cvc == -1 or cvc >= len(cols)-1:
		print '*** Error: No "CV"-Coverage column in table header line of file ',infile 
		sys.exit()
		
	############################################################################
	# loop over remaining lines to flag and complete these
	############################################################################

	while True:
		n += 1
		# read and split a single line from infile
		line = inf.readline()
		if len(line) == 0:
			# EOF
			break
		cols=line.split('\t')
		if len(cols) != ncols:
			break
			
		# compute whether this is a snp, snp-allele-coverage, total coverage
		#isnp, cn, ctot = pib_flgsnp_cov_ab (cols[cvc], cols[bgc], cols[refc], covmin, fmin)
		isnp, cn, ctot = pib_flgsnp_cov_ab (cols[cvc], cols[bgc], cols[refc])
		
		# incorporate this stuff into the line:
		outline = cols[0]
		for i in range(1,bgc):
			outline += '\t'+cols[i]
		outline += '\t%s\t%s\t%s' % (isnp, cn, ctot)
		for i in range(bgc,ncols):
			outline += '\t'+cols[i]
		# output the line (last column already has a LF):
		out.write(outline)

	inf.close()
	out.close()

##############################################
# the actual main
##############################################

nargs = len(sys.argv)
if nargs < 3:
	print pib_flgsnp.__doc__
	sys.exit()
else:
	infile = sys.argv[1]
	outfile = sys.argv[2]
#	cmin = 8
#	fmin = 0.04
#	if nargs > 3:
#		try:
#			cmin = int(sys.argv[3])
#		except:
#			print pib_flgsnp.__doc__
#			print '*** ERROR: covmin not integer. Please check your input. ***\n'
#			print 'covmin string:', sys.argv[3]
#	if nargs > 4:
#		try:
#			cmin = float(sys.argv[4])
#		except:
#			print pib_flgsnp.__doc__
#			print '*** ERROR: fmin not a number. Please check your input. ***\n'
#			print 'fin string:', sys.argv[4]
setlimits()

#pib_flgsnp (infile, outfile, cmin, fmin)
pib_flgsnp (infile, outfile)
print('Done')
