#!/usr/bin/env python
##!/usr/bin/python

import sys
import time
import resource

__author__="mforster"
__date__ ="$Feb 21, 2011 9:30 PM$"

def setlimits():
	#############################################################################
	# sets resource limits (mandatory for job interoperability on cluster)
	# to 10MB Stack + 1GB data + 1GB RLIMIT_RSS
	# first steps taken from
	# http://docs.python.org/library/resource.html and
	# http://www.doughellmann.com/PyMOTW/resource/
	# Author: MICHAEL FORSTER m.forster@ikmb.uni-kiel.de
	# Version: 13 September 2010
	#############################################################################

	#############################################################################
	# Adjust the maximum size (in bytes) of the process's heap
	#############################################################################

	soft, hard = resource.getrlimit(resource.RLIMIT_DATA)
	#print 'Data/heap limit starts as  :', soft

	# 1GB heap:
	newheap=1024*1024*1024

	resource.setrlimit(resource.RLIMIT_DATA, (newheap, hard))

	soft, hard = resource.getrlimit(resource.RLIMIT_DATA)
	#print 'Data/heap limit changed to :', soft

	#############################################################################
	# ca 10MB original stack changed to exactly 10MB stack:
	#############################################################################

	#newstack=10*1024*1024+1
	newstack=10*1024*1024
	soft, hard = resource.getrlimit(resource.RLIMIT_STACK)
	#print 'Stack limit starts as  :', soft
	resource.setrlimit(resource.RLIMIT_STACK, (newstack, hard))
	soft, hard = resource.getrlimit(resource.RLIMIT_STACK)
	#print 'Stack limit changed to :', soft

	#############################################################################
	# 1GB RLIMIT_RSS
	# "Maximale resident set size in Byte. Bei Speicherengpaessen entzieht der Kern
	# den Prozessen, die den resident set size ueberschreiten, den zuviel
	# angeforderten Speicher."
	# http://pronix.linuxdelta.de/C/Linuxprogrammierung/Linuxsystemprogrammieren_C_Kurs_Kapitel3a.shtml
	#############################################################################

	newrss=1024*1024*1024
	soft, hard = resource.getrlimit(resource.RLIMIT_RSS)
	#print 'RLIMIT_RSS starts as  :', soft
	resource.setrlimit(resource.RLIMIT_RSS, (newrss, hard))
	soft, hard = resource.getrlimit(resource.RLIMIT_RSS)
	#print 'RLIMIT_RSS changed to :', soft

def pib_tag (filetype, primerfile, infile, chr_col, pos_col, tag_col, outfile):
	'''        #######################################################################################
	#  pibase_tag.py  v1.4.3
	#######################################################################################
	#  Adds primer region tags to a pibase file (see filetype). 
	#
	#  Usage:	pibase_tag.py <typ> <pr> <in> <chr> <pos> <out> [tag]
	#  where
	#  typ:	type of input file (0=pibase_diff or pibase_annot file; others not in v1.2)
	#  pr:	primer region file, tab-separated: chr pos start end strand (int, 1-based)
	#	e.g.: 1 100 120 1
	# 	e.g.: 1 280 300 -1
	#  in:	tab-separated file, e.g. from pibase_diff or pibase_annot
	#  chr:	column number (1-based) in in-file with chromosome number 
	#  pos:	column number (1-based) in in-file with position
	#  out:	output file ( = tagged input file)
	#  tag:	(optional, default=column 2)
	#	column number, at which the 4 primer tag columns are inserted	
	#######################################################################################
	#  Author: Michael Forster, m.forster@ikmb.uni-kiel.de         Date: 22 February 2011
	#######################################################################################
	'''
	
	############################################################################
	#  read primer (chrom, start, end, strand) into list
	############################################################################

	afile = open (primerfile, 'r')
		
	# find table header
	n = -1
	while True :
		n += 1
		line = afile.readline()
		if len(line) == 0:
			# EOF
			print ' *** Error: Unable to find table header line starting with "##chr"'
			sys.exit (' *** in file %s' % (primerfile))
		if line.startswith('##chr'):
			header = line
			break
			
	# read primer regions:
	pri=[]
	
	while True :
		n += 1
		line = afile.readline()
		if len(line) == 0:
			# EOF
			break
		line = line.strip()
		cols = line.split('\t')
		try:
			if len(cols) < 4:
				# end of table
				break
			ichr = int(cols[0])
			istart = int(cols[1])-1
			iend = int(cols[2])+1
			istrand = int(cols[3])
			pri.append ( (ichr, istart, iend, istrand) )
		except:
			print ' *** Error in primer region file %s, line %s:'  % (primerfile, n)
			print ' *** line must contain 4 integer numbers'
			print ' *** chrom, start'

	afile.close()
	npri = len(pri)

	#print ' pri:'
	#print pri
	
	############################################################################
	# open infile and outfile and copy infile header lines to outfile:
	############################################################################
	
	inf = open (infile, 'r')
	out = open (outfile, 'w')
	
	outtime = time.strftime("%Y_%m_%d__%H:%M:%S") + "\n"
	out.write ('# File created by pibase_tag v1.4.3 on '+ outtime)
	out.write ('# using primer regions file: %s\n' % (primerfile))

	out.write ('# from: %s\n' %(infile))
	
	n=-1
	while True:
		n += 1
		line = inf.readline()
		if len(line) == 0:
			# EOF
			print ' *** Error: There is no table header line containing "RefSeqContext<TAB>GC[%]"'
			sys.exit (' in file %s' % (infile))
		if line.find('RefSeqContext\tGC[%]')> -1 :
			header = line
			break
		else:
			out.write(line)
	
	out.write ('# pibase_tag inserted 4 columns to tag SNPs which lie in primer regions and require special filtering:\n')
	out.write ('#  primer: 0, if SNP outside primer region +- 1 base; 1, if SNP within a primer region +- 1 base\n')
	out.write ('#  p.start: 1-based primer region start\n')
	out.write ('#  p.end: 1-based primer region end\n')
	out.write ('#  p.strand: strand for this primer (snp reads on other strand are unaffected)\n#\n')
			
	############################################################################
	# table header line for outfile
	############################################################################
	
	cols = header.split('\t')
	ncols = len(cols)
	nhcols = ncols
	outheader = ''
	if tag_col == 0:
		outheader += '#primer\tp.start\tp.end\tp.strand'
		outheader += '\t'+cols[0][1:]
		for i in range(1,ncols):
			outheader += '\t'+cols[i]
	else:
		outheader = cols[0]
		for i in range(1,tag_col):
			outheader += '\t'+cols[i]
		outheader += '\tprimer\tp.start\tp.end\tp.strand'
		for i in range(tag_col,ncols):
			outheader += '\t'+cols[i]

	out.write(outheader)

	############################################################################
	# loop over remaining lines to tag these with primer info
	############################################################################
	
	while True:
		n += 1
		line = inf.readline()
		if len(line) == 0:
			# EOF
			break
		cols=line.split('\t')
		ncols = len(cols)
		if ncols < nhcols:
			# end of table
			break
		
		try:
			chr = int (cols[chr_col])
			pos = int (cols[pos_col])
		except:
			print ' *** Error: Unable to read integer values (chrom column and pos column)'
			print ' Line:', line[:30]
			print ' cols[chr_col], cols[pos_col]: ', cols[chr_col], cols[pos_col]
			sys.exit (' Details: chrom-column %s, pos-column %s, file %s, line %s' % (chr_col+1, pos_col+1, infile, n))

		# check whether snp is in a primer region
		k = 0
		for ichr, istart, iend, istrand in pri:
			if chr == ichr and pos >= istart and pos <= iend:
				k = 1
				break
		if k == 1:
			col1 = 1
			col2 = istart+1
			col3 = iend-1
			col4 = istrand
		else:
			col1 = 0
			col2 = 0
			col3 = 0
			col4 = 0
		
		# build output line:
		if tag_col == 0:
			outline = str(col1) + '\t' + str(col2) + '\t' + str(col3) + '\t' + str(col4)
			outline += '\t'+line
		else:
			outline = cols[0]
			for i in range(1,tag_col):
				outline += '\t'+cols[i]
			outline += '\t' + str(col1) + '\t' + str(col2) + '\t' + str(col3) + '\t' + str(col4)
			for i in range(tag_col,ncols):
				outline += '\t'+cols[i]
			
		out.write(outline)
	
	out.close()
	inf.close()

##############################################
# the actual main
##############################################

if len(sys.argv) < 5:
	print pib_tag.__doc__
	sys.exit()
else:
	filetype = int(sys.argv[1])
	if filetype != 0:
		print pib_tag.__doc__
		sys.exit()
	primerfile = sys.argv[2]
	infile = sys.argv[3]
	chr_col = int(sys.argv[4])-1
	pos_col = int(sys.argv[5])-1
	outfile = sys.argv[6]
	tag_col = 1
	if len(sys.argv) == 8:
		tag_col = int(sys.argv[7])-1
	
setlimits()

pib_tag (filetype, primerfile, infile, chr_col, pos_col, tag_col, outfile)
print('Done')
