#!/usr/bin/env python
##!/usr/bin/python

import sys
import time
import resource

__author__="mforster"
__date__ ="$Jan 15, 2011 19:00 PM$"

def setlimits():
	#############################################################################
	# sets resource limits (mandatory for job interoperability on cluster)
	# to 10MB Stack + 1GB data + 1GB RLIMIT_RSS
	# first steps taken from
	# http://docs.python.org/library/resource.html and
	# http://www.doughellmann.com/PyMOTW/resource/
	# Author: MICHAEL FORSTER m.forster@ikmb.uni-kiel.de
	# Version: 13 September 2010
	#############################################################################

	#############################################################################
	# Adjust the maximum size (in bytes) of the process's heap
	#############################################################################

	soft, hard = resource.getrlimit(resource.RLIMIT_DATA)
	#print 'Data/heap limit starts as  :', soft

	# 1GB heap:
	newheap=1024*1024*1024

	resource.setrlimit(resource.RLIMIT_DATA, (newheap, hard))

	soft, hard = resource.getrlimit(resource.RLIMIT_DATA)
	#print 'Data/heap limit changed to :', soft

	#############################################################################
	# ca 10MB original stack changed to exactly 10MB stack:
	#############################################################################

	#newstack=10*1024*1024+1
	newstack=10*1024*1024
	soft, hard = resource.getrlimit(resource.RLIMIT_STACK)
	#print 'Stack limit starts as  :', soft
	resource.setrlimit(resource.RLIMIT_STACK, (newstack, hard))
	soft, hard = resource.getrlimit(resource.RLIMIT_STACK)
	#print 'Stack limit changed to :', soft

	#############################################################################
	# 1GB RLIMIT_RSS
	# "Maximale resident set size in Byte. Bei Speicherengpaessen entzieht der Kern
	# den Prozessen, die den resident set size ueberschreiten, den zuviel
	# angeforderten Speicher."
	# http://pronix.linuxdelta.de/C/Linuxprogrammierung/Linuxsystemprogrammieren_C_Kurs_Kapitel3a.shtml
	#############################################################################

	newrss=1024*1024*1024
	soft, hard = resource.getrlimit(resource.RLIMIT_RSS)
	#print 'RLIMIT_RSS starts as  :', soft
	resource.setrlimit(resource.RLIMIT_RSS, (newrss, hard))
	soft, hard = resource.getrlimit(resource.RLIMIT_RSS)
	#print 'RLIMIT_RSS changed to :', soft


def pibase_to_rdf_readfile (infile1, pmax, both) :
	############################################################################
	# Open pibase_fisherdiff file and get list of all snps
	############################################################################

	inf = open(infile1, 'r')

	#### find table header  ####
	while True :
		line = inf.readline()
		if len(line) == 0:
			# EOF
			sys.exit (' *** Error: End of file %s reached before table header could be found' % (infile1))
		if line.startswith('#D\tD_error\tp.med\tp.f4\tp.f3\tp.f2\tp.f1\tp.f0'):
			header = line
			break
	
	#### find columns p.med, A+-, B+- ####
	
	header = header.strip()
	cols=header.split('\t')
	ncols = len(cols)
	c=[]
	c.append('p.med')
	c.append('A+-')
	c.append('B+-')

	ic=[]
	maxcol=0
	for i in c:
		# print i
		n=0
		for col in cols:
			if i==col:
				# print 'found '+col+' in column '+str(n)
				ic.append(n)
				maxcol=max(maxcol,n)
				break
			n+=1
	if len(ic) != len(c):
			print ' *** Error: 1 or more columns of interest not found in header line'
			print  c
			print ic
			print ' *** file = ' + infile1
			print ' *** header line = ' + header
			sys.exit (' *** Aborted' )
	#print 'ic',ic

	#### Loop over remaining lines in input file ####
	
	snps = []
	while True :
		# read pair of lines ("-" and "+" line):
		line = inf.readline()
		if len(line) == 0:
			# EOF
			break
		cols = line.split('\t')
		if len(cols) < ncols:
			# end of table
			break
		line = inf.readline()
		if len(line) == 0:
			# EOF
			break
		cols = line.split('\t')
		if len(cols) < ncols:
			# end of table
			break
		
		p = float(cols[ic[0]])
		##print 'p = %s' % (p)
		# assume identity to first sample
		v="1"
		if p >= 0 and p < pmax:
			##print 'diff'
			# there is a significant difference to first sample
			if both == 'y':
				# both-stranded support requested:
				a = cols[ic[1]]
				b = cols[ic[2]]
				if a == '+-' and b == '+-':
					v="0"
			else:
				v="0"
		elif p == -1:
			# no coverage
			v="N"
		snps.append(v)
	inf.close()
	##print snps
	return snps
	

def pibase_to_rdf (listfile, outfile, pmax, both, elim):
	'''        #######################################################################################
	#  pibase_to_rdf v1.4.5
	#  Reads:   file of pibase_fisherdiff file names and their names in the rdf file 
	#  Outputs: binary rdf file for Network.
	#######################################################################################
	#  pibase_to_rdf  list  rdf  [[pmax] [both] [elim]]
	#  
	#  list:  file of pibase_fisherdiff files:
	#         file_name <TAB> name_in_rdf_file <LINEBREAK>
	#  rdf:   name of output file for Network
	#  pmax:  max p-value to consider (optional. default=0.01)
	#  both:  bothstranded support (y/n) (optional. default=y)
	#  elim:  eliminate "N"-columns and invariable columns (y/n) (optional. default=n)
	#######################################################################################
	#  Michael Forster, m.forster@ikmb.uni-kiel.de                            22 May 2012
	#######################################################################################
	'''

	print 'pibase_to_rdf v1.4.5'
	print 'listfile : ', listfile
	print 'rdf file : ', outfile
	print 'pmax=%s, both=%s, elim=%s' % (pmax, both, elim)
	
	############################################################################
	# Read pibase_fisherdiff file names and 
	# individual-names as they should appear in the rdf file
	############################################################################

	listf=open(listfile,'r')
	pdfiles=[]
	samples=[]
	nsamples=0
	for line in listf:
		if line=='' or line.startswith('#'):
			continue
		line=line.strip()
		cols=line.split('\t')
		# sample-name in summary table column header:
		pdfiles.append (cols[0].strip())
		if len(cols[1].strip()) > 15:
			sys.exit('*** Aborted: Individual-names must be shorter than 16 characters')
		samples.append (cols[1].strip())
		nsamples = nsamples + 1
	listf.close()
	#print 'nsamples=', nsamples
	
	if nsamples > 30000:
		print '*** Warning: There are %s individuals and Network 4.6.0.0 supports max 30000 individuals' % (nsamples)

	############################################################################
	# loop over files
	############################################################################
	
	snparray = []
	for infile1 in pdfiles:
		if infile1 == pdfiles[0]:
			# the first line contains a dummy file name, and the first sample name
			# so skip this dummy file
			continue
		snps = pibase_to_rdf_readfile (infile1, pmax, both)
		if infile1 == pdfiles[1]:
			nsnps = len(snps)
			snp2 = []
			for i in range(0,nsnps):
				snp2.append("1")
			# CAVE: Python does not append list CONTENTS unless [:] is used !!!
			snparray.append(snp2[:])
		m = len(snps) 
		if m != nsnps:
			sys.exit( '*** Aborted: In file %s the number of SNPs is %s but must be %s' % (infile1, m, nsnps) )
		snparray.append(snps[:])


	# print snparray

	############################################################################
	# Optionally identify "N"-columns and invariable columns 
	############################################################################

	# initialise vector, where 0 means don't eliminate the i'th column.
	eli=[]
	for i in range(0,nsnps):
		eli.append(0)
	
	neli = 0
	if elim == "y":
		for j in range (0,nsnps):
			# note character in the first row of the column
			i = 0
			ch = snparray [i] [j]
			# character change:
			chch = 0
			for i in range(0,nsamples):
				chj = snparray [i] [j]
				if chj == 'N':
					# an 'N' overrules any previous character change
					chch = 0
					break
				elif chj != ch:
					# character change, i.e. do not eliminate this column
					chch = 1
			if chch == 0:
				eli[j] = 1
				neli += 1
				
	#print eli
	#print snparray
	
	if nsnps-neli > 5000:
		print '*** Warning: There are %s SNPs and Network 4.6.0.0 supports max 5000 "characters" (due to be extended in a future Network version)' % (nsnps)
		print '*** Work-around for programmers: Filter away the characters which are not variable within this rdf file (i.e. "1" for all samples)'
		

	############################################################################
	# Create rdf file
	############################################################################

	out=open(outfile,'w')

	# header line:
	out.write ('  ;1.0\r\n')

	# characters line:
	line=''
	for i in range(0,nsnps):
		if eli[i] == 0:
			# line += 'SNP'+str(i+1)+';'
			line += str(i+1)+';'
	line += '\r\n'
	out.write (line)
	
	# weights line:
	line=''
	for i in range(0,nsnps):
		if eli[i] == 0:
			line += '10;'
	line += '\r\n'
	out.write (line)
	
	# samples lines:
	for i in range(0,nsamples):
		# name line:
		line = '>' + samples[i] + ';1;;;;;;;\r\n'
		out.write (line)
		# snp line:
		line = ''
		for j in range (0,nsnps):
			if eli[j] == 0:
				# line += str( snparray [j] [i] )
				line += snparray [i] [j]
		line += '\r\n'
		out.write (line)
	out.close()

##############################################
# the actual main
##############################################

n = len(sys.argv)
if  n < 3:
	print pibase_to_rdf.__doc__
	sys.exit()
else:
	listfile = sys.argv[1]
	outfile = sys.argv[2]
	pmax = 0.01
	both = "y"
	elim = "n"
	try:
		if n > 3:
			pmax = float(sys.argv[3])
		if n > 4:
			both = sys.argv[4]
		if n > 5:
			elim = sys.argv[5]
	except:
		print sys.argv
		print pibase_to_rdf.__doc__
		sys.exit()

setlimits()

pibase_to_rdf (listfile, outfile, pmax, both, elim)
print('Done')
